from __future__ import annotations

import boost_histogram as bh

import hist


def test_bh_conversion():
    h = bh.Histogram(bh.axis.Regular(3, 2, 1, metadata={"name": "x"}))
    h.axes[0].name = "y"

    h2 = hist.Hist(h)

    assert isinstance(h2.axes[0], hist.axis.Regular)
    assert h2.axes[0].name == "y"
    assert h2.axes[0].metadata == {"name": "x"}

    h3 = bh.Histogram(h2)

    assert not isinstance(h3.axes[0], hist.axis.Regular)
    assert h2.axes[0].name == "y"
    assert h3.axes[0].metadata == {"name": "x"}


def test_bh_conversion_add_metadata():
    h = bh.Histogram(bh.axis.Regular(3, 2, 1))

    h2 = hist.Hist(h, label="abc", name="def")

    assert h2.label == "abc"
    assert h2.name == "def"



from __future__ import annotations

import copy
import typing
from collections.abc import Iterator
from typing import Any

import histoprint
import numpy as np

from ._compat.typing import Self
from .axestuple import NamedAxesTuple
from .basehist import BaseHist

if typing.TYPE_CHECKING:
    import matplotlib as mpl


__all__ = ("Stack",)


class Stack:
    def __init__(
        self,
        *args: BaseHist,
    ) -> None:
        """
        Initialize Stack of histograms.
        """

        self._stack = list(args)

        if not args:
            raise ValueError("There should be histograms in the Stack")

        if not all(isinstance(a, BaseHist) for a in args):
            raise ValueError("There should be only histograms in Stack")

        first_axes = args[0].axes
        for a in args[1:]:
            if first_axes != a.axes:
                raise ValueError("The Histogram axes don't match")

    __hash__ = None  # type: ignore[assignment]

    @classmethod
    def from_iter(cls, iterable: typing.Iterable[BaseHist]) -> Self:
        """
        Create a Stack from an iterable of histograms.
        """
        return cls(*iterable)

    @classmethod
    def from_dict(cls, d: typing.Mapping[str, BaseHist]) -> Self:
        """
        Create a Stack from a dictionary of histograms. The keys of the
        dictionary are used as names.
        """

        new_dict = {k: copy.copy(h) for k, h in d.items()}
        for k, h in new_dict.items():
            h.name = k

        return cls(*new_dict.values())

    def _get_index(self, name: str | int) -> int:
        "Returns the index associated with a name. Passes through ints"
        if not isinstance(name, str):
            return name

        for n, h in enumerate(self._stack):
            if h.name == name:
                return n

        raise IndexError(f"Name not found: {name}")

    @typing.overload
    def __getitem__(self, val: int | str) -> BaseHist: ...

    @typing.overload
    def __getitem__(self, val: slice) -> Self: ...

    def __getitem__(self, val: int | slice | str) -> BaseHist | Self:
        if isinstance(val, str):
            val = self._get_index(val)
        if isinstance(val, slice):
            my_slice = slice(
                self._get_index(val.start), self._get_index(val.stop), val.step
            )
            return self.__class__(*self._stack.__getitem__(my_slice))

        return self._stack.__getitem__(val)

    def __setitem__(self, key: int | str, value: BaseHist) -> None:
        """
        Set a histogram in the Stack. Checks the axes of the histogram, they must match.
        """
        if not isinstance(value, BaseHist):
            raise ValueError("The value should be a histogram")
        if isinstance(key, str):
            key = self._get_index(key)
        if not value.axes == self._stack[key].axes:
            raise ValueError("The histogram axes don't match")

        self._stack[key] = value

    def __iter__(self) -> Iterator[BaseHist]:
        return iter(self._stack)

    def __len__(self) -> int:
        return len(self._stack)

    def __repr__(self) -> str:
        names = ", ".join(repr(getattr(h, "name", f"H{i}")) for i, h in enumerate(self))
        h = repr(self[0]) if len(self) else "Empty stack"
        return f"{self.__class__.__name__}<({names}) of {h}>"

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, Stack):
            return False
        return self._stack == other._stack

    @property
    def axes(self) -> NamedAxesTuple:
        return self._stack[0].axes

    def plot(self, *, ax: mpl.axes.Axes | None = None, **kwargs: Any) -> Any:
        """
        Plot method for Stack object.
        """

        import hist.plot

        return hist.plot.plot_stack(self, ax=ax, **kwargs)

    def show(self, **kwargs: object) -> Any:
        """
        Pretty print the stacked histograms to the console.
        """
        if "labels" not in kwargs and all(h.name is not None for h in self):
            kwargs["labels"] = [h.name for h in self]

        return histoprint.print_hist(list(self), stack=True, **kwargs)

    def __mul__(self, other: float) -> Self:
        """
        Multiply the Stack by a scalar.
        """
        return self.__class__(*(h * other for h in self))

    def __imul__(self, other: float) -> Self:
        """
        Multiply each histogram in the Stack by a scalar.
        """
        for h in self:
            h *= other  # noqa: PLW2901
        return self

    def __rmul__(self, other: float) -> Self:
        """
        Multiply the Stack by a scalar.
        """
        return self.__mul__(other)

    def __add__(self, other: float | np.typing.NDArray[Any]) -> Self:
        """
        Add a scalar or array to the Stack.
        """
        return self.__class__(*(h + other for h in self))

    def __iadd__(self, other: float | np.typing.NDArray[Any]) -> Self:
        """
        Add a scalar or array to the Stack.
        """
        for h in self:
            h += other  # noqa: PLW2901
        return self

    def __radd__(self, other: float | np.typing.NDArray[Any]) -> Self:
        """
        Add a scalar or array to the Stack.
        """
        return self.__add__(other)

    def __sub__(self, other: float | np.typing.NDArray[Any]) -> Self:
        """
        Subtract a scalar or array to the Stack.
        """
        return self.__class__(*(h - other for h in self))

    def __isub__(self, other: float | np.typing.NDArray[Any]) -> Self:
        """
        Subtract a scalar or array to the Stack.
        """
        for h in self:
            h -= other  # noqa: PLW2901
        return self

    def project(self, *args: int | str) -> Self:
        """
        Project the Stack onto a new axes.
        """
        return self.__class__(*(h.project(*args) for h in self))  # type: ignore[arg-type]


def __dir__() -> tuple[str, ...]:
    return __all__



from __future__ import annotations

import importlib.metadata

import numpy as np
import packaging.version
import pytest

import hist
from hist.serialization import from_uhi, to_uhi

bhs = pytest.importorskip("boost_histogram.serialization")

BHV = packaging.version.Version(importlib.metadata.version("boost_histogram"))
BHMETADATA = packaging.version.Version("1.6.1") <= BHV


@pytest.mark.parametrize(
    ("storage_type", "expected_type"),
    [
        pytest.param(hist.storage.AtomicInt64(), "int", id="atomic_int"),
        pytest.param(hist.storage.Int64(), "int", id="int"),
        pytest.param(
            hist.storage.Unlimited(), "double", id="unlimited"
        ),  # This always renders as double
        pytest.param(hist.storage.Double(), "double", id="double"),
    ],
)
def test_simple_to_dict(storage_type: hist.storage.Storage, expected_type: str) -> None:
    h = hist.Hist(
        hist.axis.Regular(10, 0, 1, name="x", label="X"),
        storage=storage_type,
    )
    data = to_uhi(h)

    assert data["axes"][0]["type"] == "regular"
    assert data["axes"][0]["lower"] == 0
    assert data["axes"][0]["upper"] == 1
    assert data["axes"][0]["bins"] == 10
    assert data["axes"][0]["underflow"]
    assert data["axes"][0]["overflow"]
    assert not data["axes"][0]["circular"]
    if BHMETADATA:
        assert data["axes"][0]["metadata"]["name"] == "x"
        assert data["axes"][0]["metadata"]["label"] == "X"
    assert data["storage"]["type"] == expected_type
    assert data["storage"]["values"] == pytest.approx(np.zeros(12))
    assert data["uhi_schema"] == 1
    assert "version" in data["writer_info"]["boost-histogram"]
    assert "version" in data["writer_info"]["hist"]


def test_weighed_to_dict() -> None:
    h = hist.Hist(
        hist.axis.Integer(3, 15),
        storage=hist.storage.Weight(),
    )
    data = to_uhi(h)

    assert data["axes"][0]["type"] == "regular"
    assert data["axes"][0]["lower"] == 3
    assert data["axes"][0]["upper"] == 15
    assert data["axes"][0]["bins"] == 12
    assert data["axes"][0]["underflow"]
    assert data["axes"][0]["overflow"]
    assert not data["axes"][0]["circular"]
    assert data["storage"]["type"] == "weighted"
    assert data["storage"]["values"] == pytest.approx(np.zeros(14))
    assert data["storage"]["variances"] == pytest.approx(np.zeros(14))


def test_mean_to_dict() -> None:
    h = hist.Hist(
        hist.axis.StrCategory(["one", "two", "three"]),
        storage=hist.storage.Mean(),
        name="hi",
    )
    data = to_uhi(h)

    if BHMETADATA:
        assert data["metadata"]["name"] == "hi"
    assert data["axes"][0]["type"] == "category_str"
    assert data["axes"][0]["categories"] == ["one", "two", "three"]
    assert data["axes"][0]["flow"]
    assert data["storage"]["type"] == "mean"
    assert data["storage"]["counts"] == pytest.approx(np.zeros(4))
    assert data["storage"]["values"] == pytest.approx(np.zeros(4))
    assert data["storage"]["variances"] == pytest.approx(np.zeros(4))


def test_weighted_mean_to_dict() -> None:
    h = hist.Hist(
        hist.axis.IntCategory([1, 2, 3]),
        storage=hist.storage.WeightedMean(),
    )
    h.fill([1, 2, 3, 50], weight=[10, 20, 30, 5], sample=[100, 200, 300, 1])
    h.fill([1, 2, 3, -3], weight=[10, 20, 30, 5], sample=[100, 200, 300, 1])
    data = to_uhi(h)

    assert data["axes"][0]["type"] == "category_int"
    assert data["axes"][0]["categories"] == pytest.approx([1, 2, 3])
    assert data["axes"][0]["flow"]
    assert data["storage"]["type"] == "weighted_mean"
    assert data["storage"]["sum_of_weights"] == pytest.approx(
        np.array([20, 40, 60, 10])
    )
    assert data["storage"]["sum_of_weights_squared"] == pytest.approx(
        np.array([200, 800, 1800, 50])
    )
    assert data["storage"]["values"] == pytest.approx(np.array([100, 200, 300, 1]))
    assert data["storage"]["variances"] == pytest.approx(np.zeros(4))


def test_transform_log_axis_to_dict() -> None:
    h = hist.Hist(hist.axis.Regular(10, 1, 10, transform=hist.axis.transform.log))
    data = to_uhi(h)

    assert data["axes"][0]["type"] == "variable"
    assert data["axes"][0]["edges"] == pytest.approx(
        np.exp(np.linspace(0, np.log(10), 11))
    )


def test_transform_sqrt_axis_to_dict() -> None:
    h = hist.Hist(hist.axis.Regular(10, 0, 10, transform=hist.axis.transform.sqrt))
    data = to_uhi(h)

    assert data["axes"][0]["type"] == "variable"
    assert data["axes"][0]["edges"] == pytest.approx(
        (np.linspace(0, np.sqrt(10), 11)) ** 2
    )


@pytest.mark.parametrize(
    "storage_type",
    [
        pytest.param(hist.storage.AtomicInt64(), id="atomic_int"),
        pytest.param(hist.storage.Int64(), id="int"),
        pytest.param(hist.storage.Double(), id="double"),
        pytest.param(hist.storage.Unlimited(), id="unlimited"),
    ],
)
def test_round_trip_simple(storage_type: hist.storage.Storage) -> None:
    h = hist.Hist(
        hist.axis.Regular(10, 0, 10),
        storage=storage_type,
    )
    h.fill([-1, 0, 0, 1, 20, 20, 20])
    data = to_uhi(h)
    h2 = from_uhi(data)

    if BHMETADATA and isinstance(
        storage_type, (hist.storage.Int64, hist.storage.Double)
    ):
        assert h._hist == h2._hist
        assert h == h2

    assert h.view() == pytest.approx(h2.view())


def test_round_trip_weighted() -> None:
    h = hist.Hist(
        hist.axis.Variable([1, 2, 4, 5], circular=True),
        storage=hist.storage.Weight(),
    )
    h.fill(["1", "2", "3"], weight=[10, 20, 30])
    h.fill(["1", "2", "3"], weight=[10, 20, 30])
    data = to_uhi(h)
    h2 = from_uhi(data)

    assert pytest.approx(np.array(h.axes[0])) == np.array(h2.axes[0])
    assert np.asarray(h) == pytest.approx(h2)


def test_round_trip_mean() -> None:
    h = hist.Hist(
        hist.axis.StrCategory(["1", "2", "3"]),
        storage=hist.storage.Mean(),
    )
    h.fill(["1", "2", "3"], weight=[10, 20, 30], sample=[100, 200, 300])
    h.fill(["1", "2", "3"], weight=[10, 20, 30], sample=[100, 200, 300])
    data = to_uhi(h)
    h2 = from_uhi(data)

    assert pytest.approx(np.array(h.axes[0])) == np.array(h2.axes[0])
    assert np.asarray(h) == pytest.approx(h2)


def test_round_trip_weighted_mean() -> None:
    h = hist.Hist(
        hist.axis.IntCategory([1, 2, 3]),
        storage=hist.storage.WeightedMean(),
    )
    h.fill([1, 2, 3], weight=[10, 20, 30], sample=[100, 200, 300])
    h.fill([1, 2, 3], weight=[10, 20, 30], sample=[100, 200, 300])
    data = to_uhi(h)
    h2 = from_uhi(data)

    assert pytest.approx(np.array(h.axes[0])) == np.array(h2.axes[0])
    assert np.asarray(h) == pytest.approx(h2)


def test_uhi_wrapper():
    h = hist.Hist(
        hist.axis.IntCategory([1, 2, 3]),
        storage=hist.storage.WeightedMean(),
    )
    assert to_uhi(h).keys() == h._to_uhi_().keys()
    data = h._to_uhi_()
    assert repr(from_uhi(data)) == repr(hist.Hist._from_uhi_(data))

    assert "version" in data["writer_info"]["hist"]


def test_uhi_direct_conversion():
    h = hist.Hist(
        hist.axis.IntCategory([1, 2, 3]),
        storage=hist.storage.Int64(),
    )
    uhi_dict = h._to_uhi_()
    h2 = hist.Hist(uhi_dict)
    if BHMETADATA:
        assert h == h2


def test_round_trip_native() -> None:
    h = hist.Hist(
        hist.axis.Integer(0, 10),
        storage=hist.storage.AtomicInt64(),
    )
    h.fill([-1, 0, 0, 1, 20, 20, 20])
    data = to_uhi(h)
    h2 = from_uhi(data)

    if BHMETADATA:
        assert h == h2

    assert isinstance(h2.axes[0], hist.axis.Integer)
    assert h2.storage_type is hist.storage.AtomicInt64


def test_round_trip_clean() -> None:
    h = hist.Hist(
        hist.axis.Integer(0, 10),
        storage=hist.storage.AtomicInt64(),
    )
    h.fill([-1, 0, 0, 1, 20, 20, 20])

    data = to_uhi(h)
    data = bhs.remove_writer_info(data)
    h2 = from_uhi(data)

    assert isinstance(h2.axes[0], hist.axis.Regular)
    assert h2.storage_type is hist.storage.Int64



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

from __future__ import annotations

import importlib.metadata

# -- Project information -----------------------------------------------------

project = "Hist"
copyright = "2020, Henry Schreiner"
author = "Henry Schreiner and Aman Goel and Shuo Liu"
version = importlib.metadata.version("hist")


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "matplotlib.sphinxext.plot_directive",
    "myst_parser",
    "nbsphinx",
    "sphinx.ext.autodoc",
    "sphinx.ext.intersphinx",
    "sphinx.ext.mathjax",
    "sphinx.ext.napoleon",
    "sphinx_copybutton",
    "sphinxcontrib.programoutput",
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = [
    "**.ipynb_checkpoints",
    ".DS_Store",
    ".env",
    "user-guide/notebooks/SVGHistogram.ipynb",
    "Thumbs.db",
    "_build",
    "banner_slides.md",
    "build",
]


intersphinx_mapping = {
    "python": ("https://docs.python.org/3", None),
    "boost_histogram": ("https://boost-histogram.readthedocs.io/en/latest", None),
}

# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "sphinx_book_theme"

# Config for the Sphinx book

html_baseurl = "https://hist.readthedocs.io/en/latest/"

html_theme_options = {
    "home_page_in_toc": True,
    "repository_url": "https://github.com/scikit-hep/hist",
    "use_repository_button": True,
    "use_issues_button": True,
    "use_edit_page_button": True,
}

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path: list[str] = []


# -- Options for Notebook input ----------------------------------------------

html_logo = "_images/histlogo.png"
html_title = f"Hist {version}"

nbsphinx_execute = "auto"  # auto, never

highlight_language = "python3"

nbsphinx_execute_arguments = [
    "--InlineBackend.figure_formats={'png2x'}",
    "--InlineBackend.rc=figure.dpi=96",
]

nbsphinx_kernel_name = "python3"



from __future__ import annotations

import inspect
import sys
from collections.abc import Iterable
from typing import Any, Callable, Literal, NamedTuple, Union

import numpy as np

import hist

try:
    import matplotlib.axes
    import matplotlib.pyplot as plt
    from matplotlib import patches, transforms
    from mplhep.plot import Hist1DArtists, Hist2DArtists, hist2dplot, histplot
except ModuleNotFoundError:
    print(  # noqa: T201
        "Hist requires mplhep to plot, either install hist[plot] or mplhep",
        file=sys.stderr,
    )
    raise

__all__ = (
    "hist2dplot",
    "histplot",
    "plot2d_full",
    "plot_pie",
    "plot_pull_array",
    "plot_ratio_array",
    "plot_stack",
)

_PLT_MISSING_MSG = "Hist plotting with fitting requires scipy and iminuit. Please install hist[plot,fit] or manually install dependencies."


class FitResultArtists(NamedTuple):
    line: matplotlib.lines.Line2D
    errorbar: matplotlib.container.ErrorbarContainer
    band: matplotlib.collections.PolyCollection


class RatioErrorbarArtists(NamedTuple):
    line: matplotlib.lines.Line2D
    errorbar: matplotlib.container.ErrorbarContainer


class RatioBarArtists(NamedTuple):
    line: matplotlib.lines.Line2D
    dots: matplotlib.collections.PathCollection
    bar: matplotlib.container.BarContainer


class PullArtists(NamedTuple):
    bar: matplotlib.container.BarContainer
    patch_artist: list[matplotlib.patches.Rectangle]


MainAxisArtists = Union[FitResultArtists, Hist1DArtists]

RatioArtists = Union[RatioErrorbarArtists, RatioBarArtists]
RatiolikeArtists = Union[RatioArtists, PullArtists]


def __dir__() -> tuple[str, ...]:
    return __all__


def _expand_shortcuts(key: str) -> str:
    if key == "ls":
        return "linestyle"
    return key


def _filter_dict(
    __dict: dict[str, Any], prefix: str, *, ignore: set[str] | None = None
) -> dict[str, Any]:
    """
    Keyword argument conversion: convert the kwargs to several independent args, pulling
    them out of the dict given. Prioritize prefix_kw dict.
    """

    # If passed explicitly, use that
    if f"{prefix}kw" in __dict:
        res: dict[str, Any] = __dict.pop(f"{prefix}kw")
        return {_expand_shortcuts(k): v for k, v in res.items()}

    ignore_set: set[str] = ignore or set()
    return {
        _expand_shortcuts(key[len(prefix) :]): __dict.pop(key)
        for key in list(__dict)
        if key.startswith(prefix) and key not in ignore_set
    }


def _expr_to_lambda(expr: str) -> Callable[..., Any]:
    """
    Converts a string expression like
        "a+b*np.exp(-c*x+math.pi)"
    into a callable function with 1 variable and N parameters,
        lambda x,a,b,c: "a+b*np.exp(-c*x+math.pi)"
    `x` is assumed to be the main variable, and preventing symbols
    like `foo.bar` or `foo(` from being considered as parameter.
    """
    from collections import OrderedDict
    from io import BytesIO
    from tokenize import NAME, tokenize

    varnames = []
    g = list(tokenize(BytesIO(expr.encode("utf-8")).readline))
    for ix, x in enumerate(g):
        toknum = x[0]
        tokval = x[1]
        if toknum != NAME:
            continue
        if ix > 0 and g[ix - 1][1] in {"."}:
            continue
        if ix < len(g) - 1 and g[ix + 1][1] in {".", "("}:
            continue
        varnames.append(tokval)
    varnames = list(OrderedDict.fromkeys([name for name in varnames if name != "x"]))
    lambdastr = f"lambda x,{','.join(varnames)}: {expr}"
    # pylint: disable-next=eval-used
    return eval(lambdastr)  # type: ignore[no-any-return]


def _curve_fit_wrapper(
    func: Callable[..., Any],
    xdata: np.typing.NDArray[Any],
    ydata: np.typing.NDArray[Any],
    yerr: np.typing.NDArray[Any],
    likelihood: bool = False,
) -> tuple[tuple[float, ...], np.typing.NDArray[Any]]:
    """
    Wrapper around `scipy.optimize.curve_fit`. Initial parameters (`p0`)
    can be set in the function definition with defaults for kwargs
    (e.g., `func = lambda x,a=1.,b=2.: x+a+b`, will feed `p0 = [1.,2.]` to `curve_fit`)
    """
    try:
        from scipy.optimize import curve_fit, minimize
    except ModuleNotFoundError:
        print(_PLT_MISSING_MSG, file=sys.stderr)  # noqa: T201
        raise

    params = list(inspect.signature(func).parameters.values())
    p0 = [
        1 if arg.default is inspect.Parameter.empty else arg.default
        for arg in params[1:]
    ]

    mask = yerr != 0.0
    popt, pcov = curve_fit(
        func,
        xdata[mask],
        ydata[mask],
        sigma=yerr[mask],
        absolute_sigma=True,
        p0=p0,
    )
    if likelihood:
        try:
            from iminuit import Minuit
        except ModuleNotFoundError:
            print(_PLT_MISSING_MSG, file=sys.stderr)  # noqa: T201
            raise
        from scipy.special import gammaln

        def fnll(v: Iterable[np.typing.NDArray[Any]]) -> float:
            ypred = func(xdata, *v)
            if (ypred <= 0.0).any():
                return 1e6
            return (  # type: ignore[no-any-return]
                ypred.sum() - (ydata * np.log(ypred)).sum() + gammaln(ydata + 1).sum()
            )

        # Seed likelihood fit with chi2 fit parameters
        res = minimize(fnll, popt, method="BFGS")
        popt = res.x

        # Better hessian from hesse, seeded with scipy popt
        m = Minuit(fnll, popt)
        m.errordef = 0.5
        m.hesse()
        pcov = np.array(m.covariance)
    return tuple(popt), pcov


def _plot_keywords_wrapper(ax: matplotlib.axes.Axes, legend: bool | None) -> None:
    """
    Pandas-like wrapper that wrap several useful mpl keyword arguments.
    """
    # Todo: more keywords here
    if legend:
        if ax.get_legend_handles_labels()[0]:
            ax.legend()
        else:
            raise ValueError("No labels to legend")


def plot2d_full(
    self: hist.BaseHist,
    *,
    ax_dict: dict[str, matplotlib.axes.Axes] | None = None,
    **kwargs: Any,
) -> tuple[Hist2DArtists, Hist1DArtists, Hist1DArtists]:
    """
    Plot2d_full method for BaseHist object.

    Pass a dict of axes to ``ax_dict``, otherwise, the current figure will be used.

    Other Parameters
    ----------------

    main_* : optional
        Keywords arguments prefixed with ''main_'' are forwarded to the main 2D histogram
        plot and controls its appearance.

    top_* : optional
        Keywords arguments prefixed with ''top_'' are forwarded to the top histogram plot and controls
        its appearance.

    side_* : optional
        Keywords arguments prefixed with ''side_'' are forwarded to the side histogram plot
        and controls its appearance.
    """
    # Type judgement
    if self.ndim != 2:
        raise TypeError("Only 2D-histogram has plot2d_full")

    if ax_dict is None:
        ax_dict = {}

    # Default Figure: construct the figure and axes
    if ax_dict:
        try:
            main_ax = ax_dict["main_ax"]
            top_ax = ax_dict["top_ax"]
            side_ax = ax_dict["side_ax"]
        except KeyError as err:
            raise ValueError("All axes should be all given or none at all") from err

    else:
        fig = plt.gcf()

        grid = fig.add_gridspec(
            2, 2, hspace=0, wspace=0, width_ratios=[4, 1], height_ratios=[1, 4]
        )
        main_ax = fig.add_subplot(grid[1, 0])
        top_ax = fig.add_subplot(grid[0, 0], sharex=main_ax)
        side_ax = fig.add_subplot(grid[1, 1], sharey=main_ax)

    # keyword arguments
    main_kwargs = _filter_dict(kwargs, "main_", ignore={"main_cbar"})
    top_kwargs = _filter_dict(kwargs, "top_")
    side_kwargs = _filter_dict(kwargs, "side_")

    # judge whether some arguments left
    if kwargs:
        raise ValueError(f"{set(kwargs)} not needed")

    # Plot: plot the 2d-histogram

    # main plot
    main_art = hist2dplot(self, ax=main_ax, cbar=False, **main_kwargs)

    # top plot
    top_art = histplot(
        self.project(self.axes[0].name or 0),
        ax=top_ax,
        **top_kwargs,
    )

    top_ax.spines["top"].set_visible(False)
    top_ax.spines["right"].set_visible(False)
    top_ax.xaxis.set_visible(False)

    top_ax.set_ylabel("Counts")

    # side plot
    base = side_ax.transData
    rot = transforms.Affine2D().rotate_deg(90).scale(-1, 1)

    side_art = histplot(
        self.project(self.axes[1].name or 1),
        ax=side_ax,
        transform=rot + base,
        **side_kwargs,
    )

    side_ax.spines["top"].set_visible(False)
    side_ax.spines["right"].set_visible(False)
    side_ax.yaxis.set_visible(False)
    side_ax.set_xlabel("Counts")

    return main_art, top_art, side_art


def _construct_gaussian_callable(
    __hist: hist.BaseHist,
) -> Callable[[np.typing.NDArray[Any]], np.typing.NDArray[Any]]:
    x_values = __hist.axes[0].centers
    hist_values = __hist.values()

    # gaussian with reasonable initial guesses for parameters
    constant = float(hist_values.max())
    mean = (hist_values * x_values).sum() / hist_values.sum()
    sigma = (hist_values * np.square(x_values - mean)).sum() / hist_values.sum()

    # gauss is a closure that will get evaluated in _fit_callable_to_hist
    def gauss(
        x: np.typing.NDArray[Any],
        constant: float = constant,
        mean: float = mean,
        sigma: float = sigma,
    ) -> np.typing.NDArray[Any]:
        # Note: Force np.typing.NDArray[Any] type as numpy ufuncs have type "Any"
        ret: np.typing.NDArray[Any] = constant * np.exp(
            -np.square(x - mean) / (2 * np.square(sigma))
        )
        return ret

    return gauss


def _fit_callable_to_hist(
    model: Callable[[np.typing.NDArray[Any]], np.typing.NDArray[Any]],
    histogram: hist.BaseHist,
    likelihood: bool = False,
) -> tuple[
    np.typing.NDArray[Any],
    np.typing.NDArray[Any],
    np.typing.NDArray[Any],
    tuple[tuple[float, ...], np.typing.NDArray[Any]],
]:
    """
    Fit a model, a callable function, to the histogram values.
    """
    variances = histogram.variances()
    if variances is None:
        raise RuntimeError(
            "Cannot compute from a variance-less histogram, try a Weight storage"
        )
    hist_uncert = np.sqrt(variances)

    # Infer best fit model parameters and covariance matrix
    xdata = histogram.axes[0].centers
    popt, pcov = _curve_fit_wrapper(
        model, xdata, histogram.values(), hist_uncert, likelihood=likelihood
    )
    model_values = model(xdata, *popt)

    if np.isfinite(pcov).all():
        n_samples = 100
        vopts = np.random.multivariate_normal(popt, pcov, n_samples)
        sampled_ydata = np.vstack([model(xdata, *vopt).T for vopt in vopts])
        model_uncert = np.nanstd(sampled_ydata, axis=0)
    else:
        model_uncert = np.zeros_like(hist_uncert)

    return model_values, model_uncert, hist_uncert, (popt, pcov)


def _plot_fit_result(
    __hist: hist.BaseHist,
    model_values: np.typing.NDArray[Any],
    model_uncert: np.typing.NDArray[Any],
    ax: matplotlib.axes.Axes,
    eb_kwargs: dict[str, Any],
    fp_kwargs: dict[str, Any],
    ub_kwargs: dict[str, Any],
) -> FitResultArtists:
    """
    Plot fit of model to histogram data
    """
    x_values = __hist.axes[0].centers
    variances = __hist.variances()
    if variances is None:
        raise RuntimeError(
            "Cannot compute from a variance-less histogram, try a Weight storage"
        )
    hist_uncert = np.sqrt(variances)

    errorbars = ax.errorbar(x_values, __hist.values(), hist_uncert, **eb_kwargs)

    # Ensure zorder draws data points above model
    line_zorder = errorbars[0].get_zorder() - 1
    (line,) = ax.plot(x_values, model_values, **fp_kwargs, zorder=line_zorder)

    # Uncertainty band for fitted function
    ub_kwargs.setdefault("color", line.get_color())
    if ub_kwargs["color"] == line.get_color():
        ub_kwargs.setdefault("alpha", 0.3)
    uncertainty_band = ax.fill_between(
        x_values,
        model_values - model_uncert,
        model_values + model_uncert,
        **ub_kwargs,
    )

    return FitResultArtists(line, errorbars, uncertainty_band)


def plot_ratio_array(
    __hist: hist.BaseHist,
    ratio: np.typing.NDArray[Any],
    ratio_uncert: np.typing.NDArray[Any],
    ax: matplotlib.axes.Axes,
    **kwargs: Any,
) -> RatioArtists:
    """
    Plot a ratio plot on the given axes

    Other Parameters
    ----------------
    central_value : float, optional
        Value of Y-axis at which to draw the horizontal reference line
        on the ration plot.
        Default is 1.0.

    uncert_draw_type : {"line", "bar"}, optional
        Defines how uncertainty is drawn on the ration plot.
        If "line", uncertainty is shown as error bars.
        If "bar", uncertainty is shown as bars.
        Default is "line".

    ylim : tuple of float, optional
        Y-axis limits for the ratio plot given as (ymin, ymax)
        If not provided, limits are automatically determined based on
        the ratio values and their uncertainties.

    """
    x_values = __hist.axes[0].edges[:-1]
    left_edge = __hist.axes.edges[0][0]
    right_edge = __hist.axes.edges[-1][-1]

    # Set 0 and inf to nan to hide during plotting
    ratio[ratio == 0] = np.nan
    ratio[np.isinf(ratio)] = np.nan

    central_value = kwargs.pop("central_value", 1.0)
    central_value_artist = ax.axhline(
        central_value, color="black", linestyle="dashed", linewidth=1.0
    )

    # Type now due to control flow
    axis_artists: RatioErrorbarArtists | RatioBarArtists

    uncert_draw_type = kwargs.pop("uncert_draw_type", "line")
    if uncert_draw_type == "line":
        errorbar_artists = ax.errorbar(
            x_values,
            ratio,
            yerr=ratio_uncert,
            color="black",
            marker="o",
            linestyle="none",
        )
        axis_artists = RatioErrorbarArtists(central_value_artist, errorbar_artists)
    elif uncert_draw_type == "bar":
        bar_width = (right_edge - left_edge) / len(ratio)

        bar_top = ratio + ratio_uncert[1]
        bar_bottom = ratio - ratio_uncert[0]
        # bottom can't be nan
        bar_bottom[np.isnan(bar_bottom)] = 0
        bar_height = bar_top - bar_bottom

        _ratio_points = ax.scatter(x_values, ratio, color="black")

        # Ensure zorder draws data points above uncertainty bars
        bar_zorder = _ratio_points.get_zorder() - 1
        bar_artists = ax.bar(
            x_values,
            height=bar_height,
            width=bar_width,
            bottom=bar_bottom,
            fill=False,
            linewidth=0,
            edgecolor="gray",
            hatch=3 * "/",
            zorder=bar_zorder,
        )
        axis_artists = RatioBarArtists(central_value_artist, _ratio_points, bar_artists)

    ratio_ylim = kwargs.pop("ylim", None)
    if ratio_ylim is None:
        # plot centered around central value with a scaled view range
        # the value _with_ the uncertainty in view is important so base
        # view range on extrema of value +/- uncertainty
        valid_ratios_idx = np.where(~np.isnan(ratio))
        valid_ratios = ratio[valid_ratios_idx]
        extrema = np.array(
            [
                valid_ratios - ratio_uncert[0][valid_ratios_idx],
                valid_ratios + ratio_uncert[1][valid_ratios_idx],
            ]
        )
        max_delta: float = np.amax(np.abs(extrema - central_value))
        ratio_extrema = np.abs(max_delta + central_value)

        _alpha = 2.0
        scaled_offset = max_delta + (max_delta / (_alpha * ratio_extrema))
        ratio_ylim = [central_value - scaled_offset, central_value + scaled_offset]

    ax.set_xlim(left_edge, right_edge)
    ax.set_ylim(bottom=ratio_ylim[0], top=ratio_ylim[1])

    ax.set_xlabel(__hist.axes[0].label)
    ax.set_ylabel(kwargs.pop("ylabel", "Ratio"))

    return axis_artists


def plot_pull_array(
    __hist: hist.BaseHist,
    pulls: np.typing.NDArray[Any],
    ax: matplotlib.axes.Axes,
    bar_kwargs: dict[str, Any],
    pp_kwargs: dict[str, Any],
) -> PullArtists:
    """
    Plot a pull plot on the given axes
    """
    x_values = __hist.axes[0].centers
    left_edge = __hist.axes.edges[0][0]
    right_edge = __hist.axes.edges[-1][-1]

    # Pull: plot the pulls using Matplotlib bar method
    width = (right_edge - left_edge) / len(pulls)
    bar_artists = ax.bar(x_values, pulls, width=width, **bar_kwargs)

    pp_num = pp_kwargs.pop("num", 5)
    patch_height = max(np.abs(pulls)) / pp_num
    patch_width = width * len(pulls)
    patch_artists = []
    for i in range(pp_num):
        # gradient color patches
        if "alpha" in pp_kwargs:
            pp_kwargs["alpha"] *= np.power(0.618, i)
        else:
            pp_kwargs["alpha"] = 0.5 * np.power(0.618, i)

        upRect_startpoint = (left_edge, i * patch_height)
        upRect = patches.Rectangle(
            upRect_startpoint, patch_width, patch_height, **pp_kwargs
        )
        ax.add_patch(upRect)
        downRect_startpoint = (left_edge, -(i + 1) * patch_height)
        downRect = patches.Rectangle(
            downRect_startpoint, patch_width, patch_height, **pp_kwargs
        )
        ax.add_patch(downRect)
        patch_artists.append(downRect)
        patch_artists.append(upRect)

    ax.set_xlim(left_edge, right_edge)

    ax.set_xlabel(__hist.axes[0].label)
    ax.set_ylabel("Pull")

    return PullArtists(bar_artists, patch_artists)


def _plot_ratiolike(
    self: hist.BaseHist,
    other: hist.BaseHist
    | Callable[[np.typing.NDArray[Any]], np.typing.NDArray[Any]]
    | str,
    likelihood: bool = False,
    *,
    ax_dict: dict[str, matplotlib.axes.Axes] | None = None,
    view: Literal["ratio", "pull"],
    fit_fmt: str | None = None,
    **kwargs: Any,
) -> tuple[MainAxisArtists, RatiolikeArtists]:
    r"""
    Plot ratio-like plots (ratio plots and pull plots) for BaseHist

    ``fit_fmt`` can be a string such as ``r"{name} = {value:.3g} $\pm$ {error:.3g}"``

    Other Parameters
    ----------------
    rp_* : optional
        Keyword arguments prefixed with ``rp_`` are forwarded to
        function :`plot_ratio_array` and control the appearance and behavior
        of the ratio plot (y-axis limits and the drawing style of uncertainty).

    eb_* : optional
        Keyword arguments prefixed with ``eb_`` are forwarded to
        Matplotlib error bar plotting functions and controls the
        appearance of histogram error bars.

    fp_* : optional
        Keyword arguments prefixed with ``fp_`` controls the appearance
        of model curve drawn on the main plot.

    ub_* : optional
        Keyword arguments prefixed with ``ub_`` controls the appearance
        of uncertainty bands drawn around model curve.

    bar_* : optional
        Keyword arguments prefixed with ``bar_`` are forwarded to
        Matplotlib bar plotting functions used in pull plots.

    pp_* : optional
        Keyword arguments prefixed with ``pp_`` control the appearance
        of patch-based elements used in pull plots.

    """
    from .intervals import ratio_uncertainty

    if self.ndim != 1:
        raise TypeError(
            f"Only 1D-histogram supports ratio plot, try projecting {self.__class__.__name__} to 1D"
        )
    if isinstance(other, hist.hist.Hist) and other.ndim != 1:
        raise TypeError(
            f"Only 1D-histogram supports ratio plot, try projecting other={other.__class__.__name__} to 1D"
        )

    if ax_dict:
        try:
            main_ax = ax_dict["main_ax"]
            subplot_ax = ax_dict[f"{view}_ax"]
        except KeyError as err:
            raise ValueError("All axes should be all given or none at all") from err
    else:
        fig = plt.gcf()
        grid = fig.add_gridspec(2, 1, hspace=0, height_ratios=[3, 1])

        main_ax = fig.add_subplot(grid[0])
        subplot_ax = fig.add_subplot(grid[1], sharex=main_ax)
        plt.setp(main_ax.get_xticklabels(), visible=False)

    # Keyword Argument Conversion: convert the kwargs to several independent args
    # error bar keyword arguments
    eb_kwargs = _filter_dict(kwargs, "eb_")
    eb_kwargs.setdefault("label", "Histogram Data")
    # Use "fmt" over "marker" to avoid UserWarning on keyword precedence
    eb_kwargs.setdefault("fmt", "o")
    eb_kwargs.setdefault("linestyle", "none")

    # fit plot keyword arguments
    fp_kwargs = _filter_dict(kwargs, "fp_")
    fp_kwargs.setdefault("label", "Counts")

    # bar plot keyword arguments
    bar_kwargs = _filter_dict(kwargs, "bar_", ignore={"bar_width"})

    # uncertainty band keyword arguments
    ub_kwargs = _filter_dict(kwargs, "ub_")
    ub_kwargs.setdefault("label", "Uncertainty")

    # ratio plot keyword arguments
    rp_kwargs = _filter_dict(kwargs, "rp_")
    rp_kwargs.setdefault("uncertainty_type", "poisson")
    rp_kwargs.setdefault("legend_loc", "best")
    rp_kwargs.setdefault("num_label", None)
    rp_kwargs.setdefault("denom_label", None)
    if rp_kwargs["uncertainty_type"] == "efficiency":
        rp_kwargs.setdefault("ylabel", "Efficiency")
        rp_kwargs.setdefault("ylim", [0, 1.1])

    # patch plot keyword arguments
    pp_kwargs = _filter_dict(kwargs, "pp_")

    # Judge whether some arguments are left
    if kwargs:
        raise ValueError(f"{set(kwargs)}' not needed")

    main_ax.set_ylabel(fp_kwargs["label"])

    # Computation and Fit
    hist_values = self.values()

    main_ax_artists: MainAxisArtists  # Type now due to control flow
    if callable(other) or isinstance(other, str):
        if isinstance(other, str):
            if other in {"gauss", "gaus", "normal"}:
                other = _construct_gaussian_callable(self)
            else:
                other = _expr_to_lambda(other)

        (
            compare_values,
            model_uncert,
            hist_values_uncert,
            bestfit_result,
        ) = _fit_callable_to_hist(other, self, likelihood)

        if fit_fmt is not None:
            parnames = list(inspect.signature(other).parameters)[1:]
            popt, pcov = bestfit_result
            perr = np.sqrt(np.diagonal(pcov))

            fp_label = "Fit"
            for name, value, error in zip(parnames, popt, perr):
                fp_label += "\n  "
                fp_label += fit_fmt.format(name=name, value=value, error=error)
            fp_kwargs["label"] = fp_label
        else:
            fp_kwargs["label"] = "Fitted value"

        main_ax_artists = _plot_fit_result(
            self,
            model_values=compare_values,
            model_uncert=model_uncert,
            ax=main_ax,
            eb_kwargs=eb_kwargs,
            fp_kwargs=fp_kwargs,
            ub_kwargs=ub_kwargs,
        )
    else:
        compare_values = other.values()

        self_artists = histplot(self, ax=main_ax, label=rp_kwargs["num_label"])
        other_artists = histplot(other, ax=main_ax, label=rp_kwargs["denom_label"])

        main_ax_artists = self_artists, other_artists

    subplot_ax_artists: RatiolikeArtists  # Type now due to control flow
    # Compute ratios: containing no INF values
    with np.errstate(divide="ignore", invalid="ignore"):
        if view == "ratio":
            ratios = hist_values / compare_values
            ratio_uncert = ratio_uncertainty(
                num=hist_values,
                denom=compare_values,
                uncertainty_type=rp_kwargs["uncertainty_type"],
            )
            # ratio: plot the ratios using Matplotlib errorbar or bar
            subplot_ax_artists = plot_ratio_array(
                self, ratios, ratio_uncert, ax=subplot_ax, **rp_kwargs
            )

        elif view == "pull":
            pulls: np.typing.NDArray[Any] = (
                hist_values - compare_values
            ) / hist_values_uncert

            pulls[np.isnan(pulls) | np.isinf(pulls)] = 0

            # Pass dicts instead of unpacking to avoid conflicts
            subplot_ax_artists = plot_pull_array(
                self, pulls, ax=subplot_ax, bar_kwargs=bar_kwargs, pp_kwargs=pp_kwargs
            )

    if main_ax.get_legend_handles_labels()[0]:  # Don't plot an empty legend
        main_ax.legend(loc=rp_kwargs["legend_loc"])

    return main_ax_artists, subplot_ax_artists


def get_center(x: str | int | tuple[float, float]) -> str | float:
    return (x[0] + x[1]) / 2 if isinstance(x, tuple) else x


def plot_pie(
    self: hist.BaseHist,
    *,
    ax: matplotlib.axes.Axes | None = None,
    **kwargs: Any,
) -> Any:
    if ax is None:
        fig = plt.gcf()
        ax = fig.add_subplot(111)

    data = self.density()

    labels = [str(get_center(x)) for x in self.axes[0]]

    return ax.pie(data, labels=labels, **kwargs)


def plot_stack(
    self: hist.stack.Stack,
    *,
    ax: matplotlib.axes.Axes | None = None,
    legend: bool | None = False,
    **kwargs: Any,
) -> Any:
    if self[0].ndim != 1:
        raise NotImplementedError("Please project to 1D before calling plot")

    if "label" not in kwargs and all(h.name is not None for h in self):
        kwargs["label"] = [h.name for h in self]

    ret = histplot(list(self), ax=ax, **kwargs)
    final_ax = ret[0].stairs.axes
    _plot_keywords_wrapper(final_ax, legend=legend)

    return ret



from __future__ import annotations

import warnings
from collections.abc import Iterable
from typing import Any

from boost_histogram.axis import ArrayTuple, AxesTuple

from ._compat.builtins import zip

__all__ = ("ArrayTuple", "AxesTuple", "NamedAxesTuple")


def __dir__() -> tuple[str, ...]:
    return __all__


class NamedAxesTuple(AxesTuple):
    __slots__ = ()

    def _get_index_by_name(self, name: int | str | None) -> int | None:
        if not isinstance(name, str):
            return name

        for i, ax in enumerate(self):
            if ax.name == name:
                return i
        raise KeyError(f"{name} not found in axes")

    def __getitem__(self, item: Any) -> Any:
        if isinstance(item, slice):
            item = slice(
                self._get_index_by_name(item.start),
                self._get_index_by_name(item.stop),
                self._get_index_by_name(item.step),
            )
        else:
            item = self._get_index_by_name(item)

        return super().__getitem__(item)

    @property
    def name(self) -> tuple[str]:
        """
        The names of the axes. May be empty strings.
        """
        return tuple(ax.name for ax in self)

    @name.setter
    def name(self, values: Iterable[str]) -> None:
        # strict = True from Python 3.10
        for ax, val in zip(self, values, strict=True):
            ax._raw_metadata["name"] = val

        disallowed_names = {"weight", "sample", "threads"}
        for ax in self:
            if ax.name in disallowed_names:
                disallowed_warning = (
                    f"{ax.name} is a protected keyword and cannot be used as axis name"
                )
                warnings.warn(disallowed_warning, stacklevel=2)

        valid_names = [ax.name for ax in self if ax.name]
        if len(valid_names) != len(set(valid_names)):
            raise KeyError(
                f"{self.__class__.__name__} instance cannot contain axes with duplicated names"
            )

    @property
    def label(self) -> tuple[str]:
        """
        The labels of the axes. Defaults to name if label not given, or Axis N
        if neither was given.
        """
        return tuple(ax.label for ax in self)



from __future__ import annotations

import matplotlib as mpl
import pytest

mpl.use("Agg")

mpl.use("Agg")

from hist import Hist, NamedHist
from hist.hist import BaseHist

try:
    import hist.dask as dah
except ImportError:
    dah = None


@pytest.fixture(params=[Hist, BaseHist, NamedHist])
def named_hist(request):
    return request.param


@pytest.fixture(params=[Hist, BaseHist])
def unnamed_hist(request):
    return request.param


dask_params_named = []
dask_params_unnamed = []
if dah is not None:
    dask_params_named = [dah.Hist, dah.NamedHist]
    dask_params_unnamed = [dah.Hist]


@pytest.fixture(params=dask_params_named)
def named_dask_hist(request):
    return request.param


@pytest.fixture(params=dask_params_unnamed)
def unnamed_dask_hist(request):
    return request.param



from __future__ import annotations

import ctypes
import math

import boost_histogram as bh
import numpy as np
import pytest

from hist import NamedHist, axis

# ToDo: specify what error is raised


def test_named_init():
    """
    Test named init -- whether NamedHist can be properly initialized.
    """

    # basic
    h = NamedHist(
        axis.Regular(10, 0, 1, name="x"), axis.Regular(10, 0, 1, name="y")
    ).fill(x=[0.35, 0.35, 0.45], y=[0.35, 0.35, 0.45])

    for idx in range(10):
        if idx == 3:
            assert h[idx, idx] == 2
            assert h[{"x": idx, "y": idx}] == 2
        elif idx == 4:
            assert h[idx, idx] == 1
        else:
            assert h[idx, idx] == 0
        with pytest.raises(Exception):
            h[{0: idx, 1: idx}]
    # with named axes
    assert NamedHist(
        axis.Regular(50, -3, 3, name="x"), axis.Regular(50, -3, 3, name="y")
    ).fill(x=np.random.randn(10), y=np.random.randn(10))

    assert NamedHist(axis.Boolean(name="x"), axis.Boolean(name="y")).fill(
        y=[True, False, True], x=[True, False, True]
    )

    assert NamedHist(
        axis.Variable(range(-3, 3), name="x"), axis.Variable(range(-3, 3), name="y")
    ).fill(x=np.random.randn(10), y=np.random.randn(10))

    assert NamedHist(axis.Integer(-3, 3, name="x"), axis.Integer(-3, 3, name="y")).fill(
        x=np.random.randint(-3, 3, 10), y=np.random.randint(-3, 3, 10)
    )

    assert NamedHist(
        axis.IntCategory(range(-3, 3), name="x"),
        axis.IntCategory(range(-3, 3), name="y"),
    ).fill(x=np.random.randint(-3, 3, 10), y=np.random.randint(-3, 3, 10))

    assert NamedHist(
        axis.StrCategory(["F", "T"], name="x"), axis.StrCategory("FT", name="y")
    ).fill(y=["T", "F", "T"], x=["T", "F", "T"])

    # cannot access via index
    h = NamedHist(axis.Regular(10, 0, 1, name="x")).fill(x=[0.35, 0.35, 0.45])

    for idx in range(10):
        with pytest.raises(Exception):
            h[{0: idx}]

    # with no-named axes
    with pytest.raises(Exception):
        NamedHist(axis.Regular(50, -3, 3), axis.Regular(50, -3, 3)).fill(
            x=np.random.randn(10), y=np.random.randn(10)
        )

    with pytest.raises(Exception):
        NamedHist(axis.Boolean(), axis.Boolean()).fill(
            y=[True, False, True], x=[True, False, True]
        )

    with pytest.raises(Exception):
        NamedHist(axis.Variable(range(-3, 3)), axis.Variable(range(-3, 3))).fill(
            x=np.random.randn(10), y=np.random.randn(10)
        )

    with pytest.raises(Exception):
        NamedHist(axis.Integer(-3, 3), axis.Integer(-3, 3)).fill(
            x=np.random.randn(10), y=np.random.randn(10)
        )

    with pytest.raises(Exception):
        NamedHist(
            axis.IntCategory(range(-3, 3)),
            axis.IntCategory(range(-3, 3)),
        ).fill(x=np.random.randn(10), y=np.random.randn(10))

    with pytest.raises(Exception):
        NamedHist(axis.StrCategory(["F", "T"]), axis.StrCategory("FT")).fill(
            y=["T", "F", "T"], x=["T", "F", "T"]
        )

    # with duplicated names
    with pytest.raises(Exception):
        NamedHist(axis.Regular(50, -3, 3, name="x"), axis.Regular(50, -3, 3, name="x"))

    with pytest.raises(Exception):
        NamedHist(axis.Boolean(name="y"), axis.Boolean(name="y"))

    with pytest.raises(Exception):
        NamedHist(
            axis.Variable(range(-3, 3), name="x"), axis.Variable(range(-3, 3), name="x")
        )

    with pytest.raises(Exception):
        NamedHist(axis.Integer(-3, 3, name="x"), axis.Integer(-3, 3, name="x"))

    with pytest.raises(Exception):
        NamedHist(
            axis.IntCategory(range(-3, 3), name="x"),
            axis.IntCategory(range(-3, 3), name="x"),
        )

    with pytest.raises(Exception):
        NamedHist(
            axis.StrCategory("TF", name="y"), axis.StrCategory(["T", "F"], name="y")
        )


def test_named_fill():
    """
    Test named fill -- whether NamedHist can be properly filled.
    """

    # Regular
    h = NamedHist(
        axis.Regular(10, 0, 1, name="x"),
        axis.Regular(10, 0, 1, name="y"),
        axis.Regular(2, 0, 2, name="z"),
    ).fill(
        x=[0.35, 0.35, 0.35, 0.45, 0.55, 0.55, 0.55],
        y=[0.35, 0.35, 0.45, 0.45, 0.45, 0.45, 0.45],
        z=[0, 0, 1, 1, 1, 1, 1],
    )

    z_one_only = h[{"z": bh.loc(1)}]
    for idx_x in range(10):
        for idx_y in range(10):
            if (idx_x == 3 and idx_y == 4) or (idx_x == 4 and idx_y == 4):
                assert (
                    z_one_only[idx_x, idx_y]
                    == z_one_only[{"x": idx_x, "y": idx_y}]
                    == 1
                )
            elif idx_x == 5 and idx_y == 4:
                assert (
                    z_one_only[idx_x, idx_y]
                    == z_one_only[{"x": idx_x, "y": idx_y}]
                    == 3
                )
            else:
                assert (
                    z_one_only[idx_x, idx_y]
                    == z_one_only[{"x": idx_x, "y": idx_y}]
                    == 0
                )

    # Boolean
    h = NamedHist(
        axis.Boolean(name="x"),
        axis.Boolean(name="y"),
        axis.Boolean(name="z"),
    ).fill(
        x=[True, True, True, True, True, False, True],
        y=[False, True, True, False, False, True, False],
        z=[False, False, True, True, True, True, True],
    )

    z_one_only = h[{"z": bh.loc(True)}]
    assert z_one_only[False, False] == z_one_only[{"x": False, "y": False}] == 0
    assert z_one_only[False, True] == z_one_only[{"x": False, "y": True}] == 1
    assert z_one_only[True, False] == z_one_only[{"x": True, "y": False}] == 3
    assert z_one_only[True, True] == z_one_only[{"x": True, "y": True}] == 1

    # Variable
    h = NamedHist(
        axis.Variable(range(11), name="x"),
        axis.Variable(range(11), name="y"),
        axis.Variable(range(3), name="z"),
    ).fill(
        x=[3.5, 3.5, 3.5, 4.5, 5.5, 5.5, 5.5],
        y=[3.5, 3.5, 4.5, 4.5, 4.5, 4.5, 4.5],
        z=[0, 0, 1, 1, 1, 1, 1],
    )

    z_one_only = h[{"z": bh.loc(1)}]
    for idx_x in range(10):
        for idx_y in range(10):
            if (idx_x == 3 and idx_y == 4) or (idx_x == 4 and idx_y == 4):
                assert (
                    z_one_only[idx_x, idx_y]
                    == z_one_only[{"x": idx_x, "y": idx_y}]
                    == 1
                )
            elif idx_x == 5 and idx_y == 4:
                assert (
                    z_one_only[idx_x, idx_y]
                    == z_one_only[{"x": idx_x, "y": idx_y}]
                    == 3
                )
            else:
                assert (
                    z_one_only[idx_x, idx_y]
                    == z_one_only[{"x": idx_x, "y": idx_y}]
                    == 0
                )

    # Integer
    h = NamedHist(
        axis.Integer(0, 10, name="x"),
        axis.Integer(0, 10, name="y"),
        axis.Integer(0, 2, name="z"),
    ).fill(
        x=[3, 3, 3, 4, 5, 5, 5],
        y=[3, 3, 4, 4, 4, 4, 4],
        z=[0, 0, 1, 1, 1, 1, 1],
    )

    z_one_only = h[{"z": bh.loc(1)}]
    for idx_x in range(10):
        for idx_y in range(10):
            if (idx_x == 3 and idx_y == 4) or (idx_x == 4 and idx_y == 4):
                assert (
                    z_one_only[idx_x, idx_y]
                    == z_one_only[{"x": idx_x, "y": idx_y}]
                    == 1
                )
            elif idx_x == 5 and idx_y == 4:
                assert (
                    z_one_only[idx_x, idx_y]
                    == z_one_only[{"x": idx_x, "y": idx_y}]
                    == 3
                )
            else:
                assert (
                    z_one_only[idx_x, idx_y]
                    == z_one_only[{"x": idx_x, "y": idx_y}]
                    == 0
                )

    # IntCategory
    h = NamedHist(
        axis.IntCategory(range(10), name="x"),
        axis.IntCategory(range(10), name="y"),
        axis.IntCategory(range(2), name="z"),
    ).fill(
        x=[3, 3, 3, 4, 5, 5, 5],
        y=[3, 3, 4, 4, 4, 4, 4],
        z=[0, 0, 1, 1, 1, 1, 1],
    )

    z_one_only = h[{"z": bh.loc(1)}]
    for idx_x in range(10):
        for idx_y in range(10):
            if (idx_x == 3 and idx_y == 4) or (idx_x == 4 and idx_y == 4):
                assert (
                    z_one_only[idx_x, idx_y]
                    == z_one_only[{"x": idx_x, "y": idx_y}]
                    == 1
                )
            elif idx_x == 5 and idx_y == 4:
                assert (
                    z_one_only[idx_x, idx_y]
                    == z_one_only[{"x": idx_x, "y": idx_y}]
                    == 3
                )
            else:
                assert (
                    z_one_only[idx_x, idx_y]
                    == z_one_only[{"x": idx_x, "y": idx_y}]
                    == 0
                )

    # StrCategory
    h = NamedHist(
        axis.StrCategory("FT", name="x"),
        axis.StrCategory(list("FT"), name="y"),
        axis.StrCategory(["F", "T"], name="z"),
    ).fill(
        x=["T", "T", "T", "T", "T", "F", "T"],
        y=["F", "T", "T", "F", "F", "T", "F"],
        z=["F", "F", "T", "T", "T", "T", "T"],
    )

    z_one_only = h[{"z": bh.loc("T")}]
    assert z_one_only[bh.loc("F"), bh.loc("F")] == 0
    assert z_one_only[bh.loc("F"), bh.loc("T")] == 1
    assert z_one_only[bh.loc("T"), bh.loc("F")] == 3
    assert z_one_only[bh.loc("T"), bh.loc("T")] == 1

    # without names
    with pytest.raises(Exception):
        NamedHist(
            axis.Regular(50, -3, 3, name="x"), axis.Regular(50, -3, 3, name="y")
        ).fill(np.random.randn(10), np.random.randn(10))

    with pytest.raises(Exception):
        NamedHist(axis.Boolean(name="x"), axis.Boolean(name="y")).fill(
            [True, False, True], [True, False, True]
        )

    with pytest.raises(Exception):
        NamedHist(
            axis.Variable(range(-3, 3), name="x"), axis.Variable(range(-3, 3), name="y")
        ).fill(np.random.randn(10), np.random.randn(10))

    with pytest.raises(Exception):
        NamedHist(axis.Integer(-3, 3, name="x"), axis.Integer(-3, 3, name="y")).fill(
            np.random.randn(10), np.random.randn(10)
        )

    with pytest.raises(Exception):
        NamedHist(
            axis.IntCategory(range(-3, 3), name="x"),
            axis.IntCategory(range(-3, 3), name="y"),
        ).fill(np.random.randn(10), np.random.randn(10))

    with pytest.raises(Exception):
        NamedHist(
            axis.StrCategory(["F", "T"], name="x"), axis.StrCategory("FT", name="y")
        ).fill(["T", "F", "T"], ["T", "F", "T"])

    # wrong names
    with pytest.raises(Exception):
        NamedHist(
            axis.Regular(50, -3, 3, name="x"), axis.Regular(50, -3, 3, name="y")
        ).fill(x=np.random.randn(10), z=np.random.randn(10))

    with pytest.raises(Exception):
        NamedHist(axis.Boolean(name="x"), axis.Boolean(name="y")).fill(
            y=[True, False, True], z=[True, False, True]
        )

    with pytest.raises(Exception):
        NamedHist(
            axis.Variable(range(-3, 3), name="x"), axis.Variable(range(-3, 3), name="y")
        ).fill(z=np.random.randn(10), x=np.random.randn(10))

    with pytest.raises(Exception):
        NamedHist(axis.Integer(-3, 3, name="x"), axis.Integer(-3, 3, name="y")).fill(
            x=np.random.randn(10), z=np.random.randn(10)
        )

    with pytest.raises(Exception):
        NamedHist(
            axis.IntCategory(range(-3, 3), name="x"),
            axis.IntCategory(range(-3, 3), name="y"),
        ).fill(y=np.random.randn(10), z=np.random.randn(10))

    with pytest.raises(Exception):
        NamedHist(
            axis.StrCategory(["F", "T"], name="x"), axis.StrCategory("FT", name="y")
        ).fill(z=["T", "F", "T"], x=["T", "F", "T"])

    h = NamedHist(
        axis.Regular(
            50, -4, 4, name="X", label="s [units]", underflow=False, overflow=False
        )
    ).fill(X=np.random.normal(size=10))


def test_named_access():
    """
    Test named access -- whether NamedHist bins can be accessed.
    """

    h = NamedHist(axis.Regular(10, -5, 5, name="X", label="x [units]")).fill(
        X=np.random.normal(size=1000)
    )

    assert h[6] == h[bh.loc(1)] == h[1j] == h[0j + 1] == h[-3j + 4] == h[bh.loc(1, 0)]
    h[6] = h[bh.loc(1)] = h[1j] = h[0j + 1] = h[-3j + 4] = h[bh.loc(1, 0)] = 0

    h = NamedHist(
        axis.Regular(50, -5, 5, name="Norm", label="normal distribution"),
        axis.Regular(50, -5, 5, name="Unif", label="uniform distribution"),
        axis.StrCategory(["hi", "hello"], name="Greet"),
        axis.Boolean(name="Yes"),
        axis.Integer(0, 1000, name="Int"),
    ).fill(
        Norm=np.random.normal(size=1000),
        Unif=np.random.uniform(size=1000),
        Greet=["hi"] * 800 + ["hello"] * 200,
        Yes=[True] * 600 + [False] * 400,
        Int=np.ones(1000, dtype=int),
    )

    assert h[0j, -0j + 2, "hi", True, 1]

    # mismatch dimension
    with pytest.raises(Exception):
        h[0j, -0j + 2, "hi", True]


class TestNamedStorageProxy:
    """
        Test named storage proxy suite -- whether NamedHist storage proxy \
        works properly.
    """

    def test_double(self):
        h = (
            NamedHist.new.Reg(10, 0, 1, name="x")
            .Reg(10, 0, 1, name="y")
            .Double()
            .fill(x=[0.5, 0.5], y=[0.2, 0.6])
        )

        assert h[0.5j, 0.2j] == 1
        assert h[bh.loc(0.5), bh.loc(0.6)] == 1
        assert isinstance(h[0.5j, 0.5j], float)

        # add storage to existing storage
        with pytest.raises(Exception):
            h.Double()

    def test_int64(self):
        h = NamedHist.new.Reg(10, 0, 1, name="x").Int64().fill(x=[0.5, 0.5])
        assert h[0.5j] == 2
        assert isinstance(h[0.5j], int)

        # add storage to existing storage
        with pytest.raises(Exception):
            h.Int64()

    def test_atomic_int64(self):
        h = NamedHist.new.Reg(10, 0, 1, name="x").AtomicInt64().fill(x=[0.5, 0.5])
        assert h[0.5j] == 2
        assert isinstance(h[0.5j], int)

        # add storage to existing storage
        with pytest.raises(Exception):
            h.AtomicInt64()

    def test_weight(self):
        h = NamedHist.new.Reg(10, 0, 1, name="x").Weight().fill(x=[0.5, 0.5])
        assert h[0.5j].variance == 2
        assert h[0.5j].value == 2

        # add storage to existing storage
        with pytest.raises(Exception):
            h.Weight()

    def test_mean(self):
        h = (
            NamedHist.new.Reg(10, 0, 1, name="x")
            .Mean()
            .fill(x=[0.5, 0.5], weight=[1, 1], sample=[1, 1])
        )
        assert h[0.5j].count == 2
        assert h[0.5j].value == 1
        assert h[0.5j].variance == 0

        # add storage to existing storage
        with pytest.raises(Exception):
            h.Mean()

    def test_weighted_mean(self):
        h = (
            NamedHist.new.Reg(10, 0, 1, name="x")
            .WeightedMean()
            .fill(x=[0.5, 0.5], weight=[1, 1], sample=[1, 1])
        )
        assert h[0.5j].sum_of_weights == 2
        assert h[0.5j].sum_of_weights_squared == 2
        assert h[0.5j].value == 1
        assert h[0.5j].variance == 0

        # add storage to existing storage
        with pytest.raises(Exception):
            h.WeightedMean()

    def test_unlimited(self):
        h = NamedHist.new.Reg(10, 0, 1, name="x").Unlimited().fill(x=[0.5, 0.5])
        assert h[0.5j] == 2

        # add storage to existing storage
        with pytest.raises(Exception):
            h.Unlimited()


def test_named_project():
    """
    Test named project -- whether NamedHist can be projected properly.
    """

    h = NamedHist(
        axis.Regular(
            50, -5, 5, name="A", label="a [units]", underflow=False, overflow=False
        ),
        axis.Boolean(name="B", label="b [units]"),
        axis.Variable(range(11), name="C", label="c [units]"),
        axis.Integer(0, 10, name="D", label="d [units]"),
        axis.IntCategory(range(10), name="E", label="e [units]"),
        axis.StrCategory("FT", name="F", label="f [units]"),
    )

    # via names
    assert h.project()
    assert h.project("A", "B")
    assert h.project("A", "B", "C", "D", "E", "F")

    h = NamedHist(
        axis.Regular(
            50, -5, 5, name="A", label="a [units]", underflow=False, overflow=False
        ),
        axis.Boolean(name="B", label="b [units]"),
        axis.Variable(range(11), name="C", label="c [units]"),
        axis.Integer(0, 10, name="D", label="d [units]"),
        axis.IntCategory(range(10), name="E", label="e [units]"),
        axis.StrCategory("FT", name="F", label="f [units]"),
    )

    # via indices
    with pytest.raises(Exception):
        h.project(0, 1)
    with pytest.raises(Exception):
        h.project(0, 1, 2, 3, 4, 5)

    # duplicated
    with pytest.raises(Exception):
        h.project(0, 0)

    with pytest.raises(Exception):
        h.project("A", "A")

    # wrong/mixed types
    with pytest.raises(Exception):
        h.project(2, "A")

    with pytest.raises(Exception):
        h.project(True, "A")

    # cannot found
    with pytest.raises(Exception):
        h.project(-1, 9)

    with pytest.raises(Exception):
        h.project("G", "H")


def test_named_index_access():
    """
    Test named index access -- whether NamedHist can be accessed by index.
    """

    h = NamedHist(
        axis.Regular(10, -5, 5, name="Ones"),
        axis.Regular(10, -5, 5, name="Twos"),
        axis.StrCategory(["hi", "hello"], name="Greet"),
        axis.Boolean(name="Yes"),
        axis.Integer(0, 10, name="Int"),
    ).fill(
        Ones=np.ones(10),
        Twos=np.ones(10) * 2,
        Greet=["hi"] * 8 + ["hello"] * 2,
        Yes=[True] * 6 + [False] * 4,
        Int=np.ones(10, dtype=int),
    )

    assert h[1j, 2j, "hi", True, 1] == 6
    assert (
        h[
            {
                "Ones": 6,
                "Twos": 7,
                "Greet": bh.loc("hi"),
                "Yes": bh.loc(True),
                "Int": bh.loc(1),
            }
        ]
        == 6
    )
    assert h[0j + 1, -2j + 4, "hi", True, 1] == 6
    assert (
        h[
            {
                "Ones": bh.loc(1, 0),
                "Twos": bh.loc(3, -1),
                "Greet": "hi",
                "Yes": True,
                "Int": 1,
            }
        ]
        == 6
    )

    with pytest.raises(Exception):
        h[0 : bh.loc(1, 0), 1 : bh.loc(3, -1), 2:"hi", 3:True, 4:1]

    with pytest.raises(Exception):
        h[0 : bh.loc(1, 0), 1 : bh.loc(3, -1), "Greet":"hi", 3:True, 4:1]

    assert h[0:10:2j, 0:5:5j, "hello", False, 5]
    assert len(h[::2j, 0:5, :, :, :].axes[1]) == 5
    assert len(h[:, 0:5, :, :, :].axes[1]) == 5

    # wrong loc shortcut
    with pytest.raises(Exception):
        h[0.5, 1 / 2, "hi", True, 1]

    with pytest.raises(Exception):
        h[0.5 + 1j, 1 / 2 + 1j, "hi", True, 1]

    # wrong rebin shortcut
    with pytest.raises(Exception):
        h[0:10:0.2j, 0:5:0.5j, "hello", False, 5]

    with pytest.raises(Exception):
        h[0 : 10 : 1 + 2j, 0 : 5 : 1 + 5j, "hello", False, 5]

    with pytest.raises(Exception):
        h[0:10:20j, 0:5:10j, "hello", False, 5]


def test_named_transform_proxy():
    """
    Test named transform proxy -- whether NamedHist transform proxy works properly.
    """

    h0 = NamedHist.new.Sqrt(3, 4, 25, name="x").Sqrt(4, 25, 81, name="y").Double()
    h0.fill(x=[5, 10, 17, 17], y=[26, 37, 50, 65])
    assert h0[0, 0] == 1
    assert h0[1, 1] == 1
    assert h0[2, 2] == 1
    assert h0[2, 3] == 1

    # based on existing axis
    with pytest.raises(Exception):
        NamedHist.new.Regular(3, 4, 25, name="x").Sqrt()

    # wrong value
    with pytest.raises(Exception):
        NamedHist.new.Sqrt(3, -4, 25, name="x")

    h1 = (
        NamedHist.new.Log(4, 1, 10_000, name="x")
        .Log(3, 1 / 1_000, 1, name="y")
        .Double()
    )
    h1.fill(x=[2, 11, 101, 1_001], y=[1 / 999, 1 / 99, 1 / 9, 1 / 9])
    assert h1[0, 0] == 1
    assert h1[1, 1] == 1
    assert h1[2, 2] == 1
    assert h1[3, 2] == 1

    # wrong arguments
    with pytest.raises(TypeError):
        NamedHist.new.Reg(4, 1, 10_000, name="x").Log()

    # wrong value
    with pytest.raises(ValueError):
        NamedHist.new.Log(3, -1, 10_000, name="x")

    h2 = (
        NamedHist.new.Pow(24, 1, 5, power=2, name="x")
        .Pow(124, 1, 5, power=3, name="y")
        .Double()
    )
    h2.fill(x=[1, 2, 3, 4], y=[1, 2, 3, 4])
    assert h2[0, 0] == 1
    assert h2[3, 7] == 1
    assert h2[8, 26] == 1
    assert h2[15, 63] == 1

    # based on existing axis
    with pytest.raises(TypeError):
        NamedHist.new.Reg(24, 1, 5, name="x").Pow(2)

    # wrong value
    with pytest.raises(ValueError):
        NamedHist.new.Pow(24, -1, 5, power=1 / 2, name="x")

    # lack args
    with pytest.raises(TypeError):
        NamedHist.new.Pow(24, 1, 5, name="x")

    ftype = ctypes.CFUNCTYPE(ctypes.c_double, ctypes.c_double)
    h3 = (
        NamedHist.new.Func(
            4, 1, 5, forward=ftype(math.log), inverse=ftype(math.exp), name="x"
        )
        .Func(4, 1, 5, forward=ftype(np.log), inverse=ftype(np.exp), name="y")
        .Double()
    )
    h3.fill(x=[1, 2, 3, 4], y=[1, 2, 3, 4])
    assert h3[0, 0] == 1
    assert h3[1, 1] == 1
    assert h3[2, 2] == 1
    assert h3[3, 3] == 1

    # based on existing axis
    with pytest.raises(TypeError):
        NamedHist.new.Reg(24, 1, 5, name="x").Func(ftype(math.log), ftype(math.exp))

    # Uncatchable warning
    # assert NamedHist.new.Func(
    #     4, -1, 5, name="x", forward=ftype(math.log), inverse=ftype(math.log)
    # )

    with pytest.raises(ValueError), pytest.warns(RuntimeWarning):
        NamedHist.new.Func(
            4, -1, 5, name="x", forward=ftype(np.log), inverse=ftype(np.log)
        )

    # lack args
    with pytest.raises(TypeError):
        NamedHist.new.Func(4, 1, 5, name="x")


def test_named_hist_proxy():
    """
    Test named hist proxy -- whether NamedHist hist proxy works properly.
    """

    h = NamedHist.new.Reg(10, 0, 1, name="x").Double().fill(x=[0.5, 0.5])
    assert h[0.5j] == 2

    assert type(h) is NamedHist

    assert not hasattr(NamedHist(), "new")

    h = (
        NamedHist.new.Reg(10, 0, 1, name="x")
        .Reg(10, 0, 1, name="y")
        .Double()
        .fill(x=[0.5, 0.5], y=[0.2, 0.6])
    )

    assert h[0.5j, 0.2j] == 1
    assert h[bh.loc(0.5), bh.loc(0.6)] == 1

    h = NamedHist.new.Bool(name="x").Double().fill(x=[True, True])
    assert h[bh.loc(True)] == 2

    h = (
        NamedHist.new.Bool(name="x")
        .Bool(name="y")
        .Double()
        .fill(x=[True, True], y=[True, False])
    )

    assert h[True, True] == 1
    assert h[True, False] == 1

    h = NamedHist.new.Var(range(10), name="x").Double().fill(x=[5, 5])
    assert h[5j] == 2

    h = (
        NamedHist.new.Var(range(10), name="x")
        .Var(range(10), name="y")
        .Double()
        .fill(x=[5, 5], y=[2, 6])
    )

    assert h[5j, 2j] == 1
    assert h[bh.loc(5), bh.loc(6)] == 1

    h = NamedHist.new.Int(0, 10, name="x").Double().fill(x=[5, 5])
    assert h[5j] == 2

    h = (
        NamedHist.new.Int(0, 10, name="x")
        .Int(0, 10, name="y")
        .Double()
        .fill(x=[5, 5], y=[2, 6])
    )

    assert h[5j, 2j] == 1
    assert h[bh.loc(5), bh.loc(6)] == 1

    h = NamedHist.new.IntCat(range(10), name="x").Double().fill(x=[5, 5])
    assert h[5j] == 2

    h = (
        NamedHist.new.IntCat(range(10), name="x")
        .IntCat(range(10), name="y")
        .Double()
        .fill(x=[5, 5], y=[2, 6])
    )

    assert h[5j, 2j] == 1
    assert h[bh.loc(5), bh.loc(6)] == 1

    h = NamedHist.new.StrCat("TF", name="x").Double().fill(x=["T", "T"])
    assert h["T"] == 2

    h = (
        NamedHist.new.StrCat("TF", name="x")
        .StrCat("TF", name="y")
        .Double()
        .fill(x=["T", "T"], y=["T", "F"])
    )

    assert h["T", "T"] == 1
    assert h["T", "F"] == 1


def test_named_density():
    """
    Test named density -- whether NamedHist density work properly.
    """

    for data in range(10, 20, 10):
        h = NamedHist(axis.Regular(10, -3, 3, name="x")).fill(x=np.random.randn(data))
        assert pytest.approx(sum(h.density()), 2) == pytest.approx(10 / 6, 2)


def test_named_axestuple():
    """
    Test named axes tuple -- whether NamedHist axes tuple work properly.
    """

    h = NamedHist(
        axis.Regular(20, 0, 12, name="A"),
        axis.Regular(10, 1, 3, name="B", label="Beta"),
        axis.Regular(15, 3, 5, name="C"),
        axis.Regular(5, 3, 2, name="D", label="Axis 3"),
    )

    assert h.axes.name == ("A", "B", "C", "D")
    assert h.axes.label == ("A", "Beta", "C", "Axis 3")

    assert h.axes[0].size == 20
    assert h.axes["A"].size == 20

    assert h.axes[1].size == 10
    assert h.axes["B"].size == 10

    assert h.axes[2].size == 15

    assert h.axes[:2].size == (20, 10)
    assert h.axes["A":"B"].size == (20,)
    assert h.axes[:"B"].size == (20,)
    assert h.axes["B":].size == (10, 15, 5)



from __future__ import annotations

from hist import Stack, axis


def test_1D_empty_repr(named_hist):
    h = named_hist.new.Reg(10, -1, 1, name="x", label="y").Double()
    html = h._repr_html_()
    assert html
    assert "name='x'" in repr(h)
    assert "label='y'" in repr(h)


def test_1D_var_empty_repr(named_hist):
    h = named_hist.new.Var(range(10), name="x", label="y").Double()
    html = h._repr_html_()
    assert html
    assert "name='x'" in repr(h)
    assert "label='y'" in repr(h)


def test_1D_int_empty_repr(named_hist):
    h = named_hist.new.Int(-9, 9, name="x", label="y").Double()
    html = h._repr_html_()
    assert html
    assert "name='x'" in repr(h)
    assert "label='y'" in repr(h)


def test_1D_intcat_empty_repr(named_hist):
    h = named_hist.new.IntCat([1, 3, 5], name="x", label="y").Double()
    html = h._repr_html_()
    assert html
    assert "name='x'" in repr(h)
    assert "label='y'" in repr(h)


def test_1D_strcat_empty_repr(named_hist):
    h = named_hist.new.StrCat(["1", "3", "5"], name="x", label="y").Double()
    html = h._repr_html_()
    assert html
    assert "name='x'" in repr(h)
    assert "label='y'" in repr(h)


def test_2D_empty_repr(named_hist):
    h = (
        named_hist.new.Reg(10, -1, 1, name="x", label="y")
        .Int(0, 15, name="p", label="q")
        .Double()
    )
    html = h._repr_html_()
    assert html
    assert "name='x'" in repr(h)
    assert "name='p'" in repr(h)
    assert "label='y'" in repr(h)
    assert "label='q'" in repr(h)


def test_1D_circ_empty_repr(named_hist):
    h = named_hist.new.Reg(10, -1, 1, circular=True, name="R", label="r").Double()
    html = h._repr_html_()
    assert html
    assert "name='R'" in repr(h)
    assert "label='r'" in repr(h)


def test_ND_empty_repr(named_hist):
    h = (
        named_hist.new.Reg(10, -1, 1, name="x", label="y")
        .Reg(12, -3, 3, name="p", label="q")
        .Reg(15, -2, 4, name="a", label="b")
        .Double()
    )
    html = h._repr_html_()
    assert html is None


def test_empty_mega_repr(named_hist):
    h = named_hist.new.Reg(1001, -1, 1, name="x").Double()
    html = h._repr_html_()
    assert html is None

    h = named_hist.new.Reg(201, -1, 1, name="x").Reg(100, 0, 1, name="y").Double()
    html = h._repr_html_()
    assert html is None


def test_stack_repr(named_hist):
    a1 = axis.Regular(
        50, -5, 5, name="A", label="a [unit]", underflow=False, overflow=False
    )
    a2 = axis.Regular(
        50, -5, 5, name="A", label="a [unit]", underflow=False, overflow=False
    )
    assert "name='A'" in repr(Stack(named_hist(a1), named_hist(a2)))
    assert "label='a [unit]'" in repr(Stack(named_hist(a1), named_hist(a2)))



from __future__ import annotations

from typing import Any, Callable

import numpy as np
from boost_histogram.axis import Axis

import hist

from .svgutils import (
    SupportsStr,
    circle,
    div,
    html,
    line,
    polygon,
    polyline,
    rect,
    svg,
    text,
)


def _desc_hist(h: hist.BaseHist) -> str:
    main_sum = h.sum()
    flow_too_sum = h.sum(flow=True)

    output = "<br/>\n".join(str(h) for h in h.axes)
    output += '<br/>\n<hr style="margin-top:.2em; margin-bottom:.2em;"/>\n'
    output += f"{h.storage_type()} "
    output += f"Σ={main_sum}"
    if main_sum != flow_too_sum:
        output += f" <em>({flow_too_sum} with flow)</em>"
    output += "\n"
    return output


def html_hist(h: hist.BaseHist, function: Callable[[hist.BaseHist], svg]) -> html:
    left_column = div(function(h), style="width:290px;")
    right_column = div(_desc_hist(h), style="flex=grow:1;")

    container = div(
        left_column, right_column, style="display:flex; align-items:center;"
    )

    return html(container)


def make_text(txt: str | float, **kwargs: SupportsStr) -> text:
    style = "fill:currentColor;"
    kwargs["style"] = style + str(kwargs.get("style", ""))
    if isinstance(txt, float):
        txt = format(txt, ".3g")
    return text(txt, text_anchor="middle", **kwargs)


def make_ax_text(ax: Axis, **kwargs: SupportsStr) -> text:
    style = "" if ax.label else "font-family: monospace;"
    kwargs["style"] = style + str(kwargs.get("style", ""))
    return make_text(ax.label or ax.name, **kwargs)


def svg_hist_1d(h: hist.BaseHist) -> svg:
    width = 250
    height = 100

    assert h.ndim == 1, "Must be 1D"
    assert not h.axes[0].traits.circular, "Must not be circular"

    (edges,) = h.axes.edges
    norm_edges = (edges - edges[0]) / (edges[-1] - edges[0])
    density = h.density()
    max_dens: float = np.amax(density) or 1  # type: ignore[redundant-expr, unreachable]
    norm_vals: np.typing.NDArray[Any] = density / max_dens

    arr: np.typing.NDArray[np.float64] = np.empty(
        (2, len(norm_vals) * 2 + 2), dtype=float
    )
    arr[0, 0:-1:2] = arr[0, 1::2] = width * norm_edges
    arr[1, 1:-2:2] = arr[1, 2:-1:2] = -height * norm_vals
    arr[1, 0] = arr[1, -1] = 0

    points = " ".join(f"{x:3g},{y:.3g}" for x, y in arr.T)
    bins = polyline(points=points, style="fill:none; stroke:currentColor;")

    ax_line = line(
        x1=-5,
        y1=0,
        x2=width + 5,
        y2=0,
        style="fill:none;stroke-width:2;stroke:currentColor",
    )

    lower = make_text(edges[0], x=0, y=15)
    upper = make_text(edges[-1], x=width, y=15)
    label = make_ax_text(
        h.axes[0],
        x=width / 2,
        y=15,
    )

    return svg(
        ax_line,
        lower,
        upper,
        label,
        bins,
        viewBox=f"-10 {-height - 5} {width + 20} {height + 20}",
    )


def svg_hist_1d_c(h: hist.BaseHist) -> svg:
    width = 250
    height = 250
    radius = 100
    inner_radius = 20

    assert h.ndim == 1, "Must be 1D"
    assert h.axes[0].traits.circular, "Must be circular"

    (edges,) = h.axes.edges
    norm_edges = (edges - edges[0]) / (edges[-1] - edges[0]) * np.pi * 2
    density = h.density()
    max_dens = np.amax(density) or 1  # type: ignore[redundant-expr, var-annotated, unreachable]
    norm_vals: np.typing.NDArray[Any] = density / max_dens

    arr: np.typing.NDArray[np.float64] = np.empty((2, len(norm_vals) * 2), dtype=float)
    arr[0, :-1:2] = arr[0, 1::2] = norm_edges[:-1]
    arr[1, :-1:2] = arr[1, 1::2] = inner_radius + norm_vals * (radius - inner_radius)
    arr[1] = np.roll(arr[1], shift=1)

    xs = arr[1] * np.cos(arr[0])
    ys = arr[1] * np.sin(arr[0])

    points = " ".join(f"{x:3g},{y:.3g}" for x, y in zip(xs, ys))
    bins = polygon(points=points, style="fill:none; stroke:currentColor;")

    center = circle(
        cx="0",
        cy="0",
        r=f"{inner_radius}",
        style="fill:none;stroke-width:2;stroke:currentColor",
    )

    return svg(bins, center, viewBox=f"{-width / 2} {-height / 2} {width} {height}")


def svg_hist_2d(h: hist.BaseHist) -> svg:
    width = 250
    height = 250
    assert h.ndim == 2, "Must be 2D"

    e0, e1 = (h.axes[0].edges, h.axes[1].edges)
    ex = (e0 - e0[0]) / (e0[-1] - e0[0]) * width
    ey = -(e1 - e1[0]) / (e1[-1] - e1[0]) * height

    density = h.density()
    max_dens = np.amax(density) or 1  # type: ignore[redundant-expr, var-annotated, unreachable]
    norm_vals: np.typing.NDArray[Any] = density / max_dens

    boxes = []
    for r, (up_edge, bottom_edge) in enumerate(zip(ey[:-1], ey[1:])):
        ht = up_edge - bottom_edge
        for c, (left_edge, right_edge) in enumerate(zip(ex[:-1], ex[1:])):
            opacity = norm_vals[c, r]
            wt = left_edge - right_edge
            boxes.append(
                rect(
                    x=left_edge,
                    y=bottom_edge,
                    width=round(abs(wt), 3),
                    height=round(abs(ht), 3),
                    opacity=opacity,
                    fill="currentColor",
                    stroke_width=0.1,
                )
            )

    texts = [
        make_text(e0[0], x=0, y=13),
        make_text(e0[-1], x=width, y=13),
        make_text(e1[0], x=-10, y=0),
        make_text(e1[-1], x=-10, y=-height),
        make_ax_text(
            h.axes[0],
            x=width / 2,
            y=13,
        ),
        make_ax_text(
            h.axes[1],
            x=-10,
            y=-height / 2,
            transform=f"rotate(-90,{-10},{-height / 2})",
        ),
    ]

    return svg(
        *texts, *boxes, viewBox=f"{-20} {-height - 20} {width + 40} {height + 40}"
    )



from __future__ import annotations

import importlib.util

if not importlib.util.find_spec("dask_histogram"):
    raise ModuleNotFoundError(
        """for hist.dask, install the 'dask_histogram' package with:
        pip install dask_histogram
        or
        conda install dask_histogram"""
    )

from .hist import Hist
from .namedhist import NamedHist

new = Hist.new

__all__ = ["Hist", "NamedHist", "new"]



from __future__ import annotations

import pytest

from hist import axis, hist


def test_axis_names():
    """
    Test axis names -- whether axis names work.
    """

    assert axis.Regular(50, -3, 3, name="x0")
    assert axis.Boolean(name="x_")
    assert axis.Variable(range(-3, 3), name="xx")
    assert axis.Integer(-3, 3, name="x_x")
    assert axis.IntCategory(range(-3, 3), name="X__X")
    assert axis.StrCategory("FT", name="X00")

    assert axis.Regular(50, -3, 3, name="")
    assert axis.Boolean(name="")
    assert axis.Variable(range(-3, 3))
    assert axis.Integer(-3, 3, name="")
    assert axis.IntCategory(range(-3, 3), name="")
    assert axis.StrCategory("FT")


def test_axis_flow():
    assert axis.Regular(9, 0, 8, flow=False) == axis.Regular(
        9, 0, 8, underflow=False, overflow=False
    )
    assert axis.Variable([1, 2, 3], flow=False) == axis.Variable(
        [1, 2, 3], underflow=False, overflow=False
    )
    assert axis.Integer(0, 8, flow=False) == axis.Integer(
        0, 8, underflow=False, overflow=False
    )

    assert axis.Regular(9, 0, 8, flow=False, underflow=True) == axis.Regular(
        9, 0, 8, overflow=False
    )
    assert axis.Variable([1, 2, 3], flow=False, underflow=True) == axis.Variable(
        [1, 2, 3], overflow=False
    )
    assert axis.Integer(0, 8, flow=False, underflow=True) == axis.Integer(
        0, 8, overflow=False
    )

    assert axis.Regular(9, 0, 8, flow=False, overflow=True) == axis.Regular(
        9, 0, 8, underflow=False
    )
    assert axis.Variable([1, 2, 3], flow=False, overflow=True) == axis.Variable(
        [1, 2, 3], underflow=False
    )
    assert axis.Integer(0, 8, flow=False, overflow=True) == axis.Integer(
        0, 8, underflow=False
    )


def test_axis_disallowed_names():
    with pytest.warns(UserWarning, match="weight is a protected keyword"):
        hist.Hist(axis.Regular(10, 0, 10, name="weight"))
    with pytest.warns(UserWarning, match="sample is a protected keyword"):
        hist.Hist(axis.Regular(10, 0, 10, name="sample"))
    with pytest.warns(UserWarning, match="threads is a protected keyword"):
        hist.Hist(axis.Regular(10, 0, 10, name="threads"))



from __future__ import annotations

import numpy as np
from pytest import approx

import hist


def test_example():
    xy = np.array(
        [
            [-2, 1.5],
            [-2, -3.5],
            [-2, 1.5],  # x = -2
            [0.0, -2.0],
            [0.0, -2.0],
            [0.0, 0.0],
            [0.0, 2.0],
            [0.0, 4.0],  # x = 0
            [2, 1.5],  # x = +2
        ]
    )
    h = hist.Hist(
        hist.axis.Regular(5, -5, 5, name="x"), hist.axis.Regular(5, -5, 5, name="y")
    ).fill(*xy.T)

    # Profile out the y-axis
    hp = h.profile("y")

    # Exclude edge bins since no values from above will fall into them
    # When there are values there, ROOT does something funky in those bins,
    # despite these bins not being in the axis that is profiled out, and
    # despite there being no overflow... to be understood.
    assert hp.values()[1:-1] == approx(np.array([0.0, 0.4, 2.0]))
    assert hp.variances()[1:-1] == approx(
        np.array([2.66666667, 1.088, float("nan")]), nan_ok=True
    )



from __future__ import annotations

import argparse
import shutil
import sys
from pathlib import Path

import boost_histogram as bh
from histoprint import print_hist

import hist


def main() -> None:
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-i", "--input", type=str, help="input file to read from (stdin by default)"
    )
    parser.add_argument("-b", "--buckets", type=int, help="number of bins", default=20)
    parser.add_argument(
        "-s",
        "--screen-width",
        type=int,
        help="maximum screen width",
        default=shutil.get_terminal_size()[0],
    )
    parser.add_argument("-t", "--label", type=str, help="label for plot")
    parser.add_argument("-o", "--output-image", type=str, help="save image to file")
    args = parser.parse_args()

    print(  # noqa: T201
        "Classic hist interface - please use histoprint instead; this supports multiple file formats and much more!"
    )

    with Path(args.input).open(encoding="utf-8") if args.input else sys.stdin as f:
        values = [float(v) for v in f]

    h = bh.numpy.histogram(values, bins=args.buckets, histogram=hist.Hist)

    if args.output_image:
        import matplotlib.pyplot as plt

        h.plot()
        plt.savefig(args.output_image)
    else:
        print_hist(
            h,
            label=args.label,
            summary=True,
        )


if __name__ == "__main__":
    main()



from __future__ import annotations

import sys
from typing import TYPE_CHECKING, Any

if sys.version_info < (3, 11):
    from typing_extensions import Self
else:
    from typing import Self

if TYPE_CHECKING:
    from numpy import ufunc as Ufunc
    from numpy.typing import ArrayLike
else:
    ArrayLike = Any
    Ufunc = Any


__all__ = ["ArrayLike", "Self", "Ufunc"]


def __dir__() -> list[str]:
    return __all__



from __future__ import annotations

import pytest

from hist import Hist


def test_axestuple():
    """
    Test axistuples -- whether axistuple setattr works.
    """

    h = Hist.new.Regular(50, -3, 3, name="X").Regular(20, -3, 3, name="Y").Double()
    h.axes.name = ("A", "B")
    h.axes.label = ("A-unit", "B-unit")

    assert h.axes[0].name == "A"
    assert h.axes[1].name == "B"
    assert h.axes[0].label == "A-unit"
    assert h.axes[1].label == "B-unit"

    with (
        pytest.warns(UserWarning, match="weight is a protected keyword"),
        pytest.warns(UserWarning, match="sample is a protected keyword"),
    ):
        h.axes.name = ("weight", "sample")

    with pytest.raises(Exception, match="cannot contain axes with duplicated names"):
        h.axes.name = ("A", "A")

    # Our backport has a simpler error message
    with pytest.raises(
        Exception,
        match=r"argument 2 is longer than argument 1|arguments are not the same length",
    ):
        h.axes.name = ("A", "B", "C")



from __future__ import annotations

import numpy as np
import pytest

import hist

da = pytest.importorskip("dask.array")
dah = pytest.importorskip("hist.dask")


@pytest.mark.parametrize("use_weights", [True, False])
def test_unnamed_5D_strcat_intcat_rectangular(unnamed_dask_hist, use_weights):
    x = da.random.standard_normal(size=(2000, 3), chunks=(400, 3))
    if use_weights:
        weights = da.random.uniform(0.5, 0.75, size=x.shape[0], chunks=x.chunksize[0])
        storage = hist.storage.Weight()
    else:
        weights = None
        storage = hist.storage.Double()

    h = unnamed_dask_hist(
        hist.axis.StrCategory([], growth=True, name="strcat"),
        hist.axis.IntCategory([], growth=True, name="intcat"),
        hist.axis.Regular(8, -3.5, 3.5, name="x"),
        hist.axis.Regular(7, -3.3, 3.3, name="y"),
        hist.axis.Regular(9, -3.2, 3.2, name="z"),
        storage=storage,
    )
    xT = x.T
    h.fill(strcat="testcat1", intcat=1, x=xT[0], y=xT[1], z=xT[2], weight=weights)
    h.fill(strcat="testcat2", intcat=2, x=xT[0], y=xT[1], z=xT[2], weight=weights)
    h = h.compute()

    control = h.__class__(*h.axes, storage=h.storage_type())
    xTc = x.compute().T
    if use_weights:
        control.fill(
            strcat="testcat1",
            intcat=1,
            x=xTc[0],
            y=xTc[1],
            z=xTc[2],
            weight=weights.compute(),
        )
        control.fill(
            strcat="testcat2",
            intcat=2,
            x=xTc[0],
            y=xTc[1],
            z=xTc[2],
            weight=weights.compute(),
        )
    else:
        control.fill(strcat="testcat1", intcat=1, x=xTc[0], y=xTc[1], z=xTc[2])
        control.fill(strcat="testcat2", intcat=2, x=xTc[0], y=xTc[1], z=xTc[2])

    assert np.allclose(h.counts(), control.counts())
    if use_weights:
        assert np.allclose(h.variances(), control.variances())

    assert len(h.axes[0]) == 2
    assert len(control.axes[0]) == 2
    assert all(cx == hx for cx, hx in zip(control.axes[0], h.axes[0]))

    assert len(h.axes[1]) == 2
    assert len(control.axes[1]) == 2
    assert all(cx == hx for cx, hx in zip(control.axes[1], h.axes[1]))


@pytest.mark.parametrize("use_weights", [True, False])
def test_named_5D_strcat_intcat_rectangular(named_dask_hist, use_weights):
    x = da.random.standard_normal(size=(2000, 3), chunks=(400, 3))
    if use_weights:
        weights = da.random.uniform(0.5, 0.75, size=x.shape[0], chunks=x.chunksize[0])
        storage = hist.storage.Weight()
    else:
        weights = None
        storage = hist.storage.Double()

    h = named_dask_hist(
        hist.axis.StrCategory([], growth=True, name="strcat"),
        hist.axis.IntCategory([], growth=True, name="intcat"),
        hist.axis.Regular(8, -3.5, 3.5, name="x"),
        hist.axis.Regular(7, -3.3, 3.3, name="y"),
        hist.axis.Regular(9, -3.2, 3.2, name="z"),
        storage=storage,
    )
    xT = x.T
    h.fill(strcat="testcat1", intcat=1, x=xT[0], y=xT[1], z=xT[2], weight=weights)
    h.fill(strcat="testcat2", intcat=2, x=xT[0], y=xT[1], z=xT[2], weight=weights)
    h = h.compute()

    control = h.__class__(*h.axes, storage=h.storage_type())
    xTc = x.compute().T
    if use_weights:
        control.fill(
            strcat="testcat1",
            intcat=1,
            x=xTc[0],
            y=xTc[1],
            z=xTc[2],
            weight=weights.compute(),
        )
        control.fill(
            strcat="testcat2",
            intcat=2,
            x=xTc[0],
            y=xTc[1],
            z=xTc[2],
            weight=weights.compute(),
        )
    else:
        control.fill(strcat="testcat1", intcat=1, x=xTc[0], y=xTc[1], z=xTc[2])
        control.fill(strcat="testcat2", intcat=2, x=xTc[0], y=xTc[1], z=xTc[2])

    assert np.allclose(h.counts(), control.counts())
    if use_weights:
        assert np.allclose(h.variances(), control.variances())

    assert len(h.axes[0]) == 2
    assert len(control.axes[0]) == 2
    assert all(cx == hx for cx, hx in zip(control.axes[0], h.axes[0]))

    assert len(h.axes[1]) == 2
    assert len(control.axes[1]) == 2
    assert all(cx == hx for cx, hx in zip(control.axes[1], h.axes[1]))



from __future__ import annotations

import fnmatch
import functools
import itertools
import operator
import typing
import warnings
from collections.abc import Generator, Iterator, Mapping, Sequence
from typing import (
    Any,
    Callable,
    Protocol,
    SupportsIndex,
    Union,
)

import boost_histogram as bh
import histoprint
import numpy as np

import hist

from . import interop
from ._compat.typing import ArrayLike, Self
from .axestuple import NamedAxesTuple
from .axis import AxisProtocol
from .quick_construct import MetaConstructor
from .storage import Storage
from .svgplots import html_hist, svg_hist_1d, svg_hist_1d_c, svg_hist_2d

if typing.TYPE_CHECKING:
    from builtins import ellipsis

    import matplotlib.axes
    from mplhep.plot import Hist1DArtists, Hist2DArtists

    from .plot import FitResultArtists, MainAxisArtists, RatiolikeArtists


class SupportsLessThan(Protocol):
    def __lt__(self, __other: Any) -> bool: ...


InnerIndexing = Union[
    SupportsIndex, str, Callable[[bh.axis.Axis], int], slice, "ellipsis"
]
IndexingWithMapping = Union[InnerIndexing, Mapping[Union[int, str], InnerIndexing]]
IndexingExpr = Union[IndexingWithMapping, tuple[IndexingWithMapping, ...]]
AxisTypes = Union[AxisProtocol, tuple[int, float, float]]


# Workaround for bug in mplhep
def _proc_kw_for_lw(kwargs: Mapping[str, Any]) -> dict[str, Any]:
    return {
        f"{k[:-3]}_linestyle"
        if k.endswith("_ls")
        else "linestyle"
        if k == "ls"
        else k: v
        for k, v in kwargs.items()
    }


def process_mistaken_quick_construct(
    axes: Sequence[AxisTypes | hist.quick_construct.ConstructProxy],
) -> Generator[AxisTypes, None, None]:
    for ax in axes:
        if isinstance(ax, hist.quick_construct.ConstructProxy):
            yield from ax.axes
        else:
            yield ax


NO_METADATA = object()


class BaseHist(bh.Histogram, metaclass=MetaConstructor, family=hist):
    __slots__ = ()

    @typing.overload
    def __init__(
        self,
        arg: dict[str, Any],
        /,
        *,
        data: np.typing.NDArray[Any] | None = ...,
        metadata: Any = ...,
        label: str | None = ...,
        name: str | None = ...,
    ) -> None: ...

    @typing.overload
    def __init__(
        self,
        arg: Self | bh.Histogram,
        /,
        *,
        data: np.typing.NDArray[Any] | None = ...,
        metadata: Any = ...,
        label: str | None = ...,
        name: str | None = ...,
    ) -> None: ...

    @typing.overload
    def __init__(
        self,
        *axes: AxisTypes | Storage | str,
        storage: Storage = ...,
        metadata: Any = ...,
        data: np.typing.NDArray[Any] | None = ...,
        label: str | None = ...,
        name: str | None = ...,
    ) -> None: ...

    def __init__(
        self,
        *in_args: Self | bh.Histogram | dict[str, Any] | AxisTypes | Storage | str,
        storage: Storage | str | None = None,
        metadata: Any = NO_METADATA,
        data: np.typing.NDArray[Any] | None = None,
        label: str | None = None,
        name: str | None = None,
    ) -> None:
        """
        Initialize BaseHist object. Axis params can contain the names.
        """
        self._hist: Any = None
        self.axes: NamedAxesTuple

        args: tuple[AxisTypes, ...]

        if in_args and storage is None and isinstance(in_args[-1], (Storage, str)):
            storage = in_args[-1]
            args = in_args[:-1]  # type: ignore[assignment]
        else:
            args = in_args  # type: ignore[assignment]

        # Support raw Quick Construct being accidentally passed in
        args = tuple(ax for ax in process_mistaken_quick_construct(args))

        if isinstance(storage, str):
            storage_str = storage.title()
            if storage_str == "Atomicint64":
                storage_str = "AtomicInt64"
            elif storage_str == "Weightedmean":
                storage_str = "WeightedMean"
            storage = getattr(bh.storage, storage_str)()
        elif isinstance(storage, type):
            msg = f"Please use '{storage.__name__}()' instead of '{storage.__name__}'"  # type: ignore[unreachable]
            warnings.warn(msg, stacklevel=2)
            storage = storage()

        if metadata is NO_METADATA:
            super().__init__(*args, storage=storage)  # type: ignore[call-overload]
        else:
            super().__init__(*args, storage=storage, metadata=metadata)  # type: ignore[call-overload]

        disallowed_names = {"weight", "sample", "threads"}
        for ax in self.axes:
            if ax.name in disallowed_names:
                disallowed_warning = (
                    f"{ax.name} is a protected keyword and cannot be used as axis name"
                )
                warnings.warn(disallowed_warning, stacklevel=2)

        valid_names = [ax.name for ax in self.axes if ax.name]
        if len(valid_names) != len(set(valid_names)):
            raise KeyError(
                f"{self.__class__.__name__} instance cannot contain axes with duplicated names"
            )
        for i, ax in enumerate(self.axes):
            # label will return name if label is not set, so this is safe
            if not ax.label:
                ax.label = f"Axis {i}"

        if data is not None:
            self[...] = data

        self.name = name
        self.label = label

    def _generate_axes_(self) -> NamedAxesTuple:
        """
        This is called to fill in the axes. Subclasses can override it if they need
        to change the axes tuple.
        """

        return NamedAxesTuple(self._axis(i) for i in range(self.ndim))

    def _repr_html_(self) -> str | None:
        if self.size == 0:
            return None

        if self.ndim == 1 and len(self.axes[0]) <= 1000:
            return str(
                html_hist(
                    self,
                    svg_hist_1d_c if self.axes[0].traits.circular else svg_hist_1d,
                )
            )

        if self.ndim == 2 and len(self.axes[0]) <= 200 and len(self.axes[1]) <= 200:
            return str(html_hist(self, svg_hist_2d))

        return None

    def _name_to_index(self, name: str) -> int:
        """
            Transform axis name to axis index, given axis name, return axis \
            index.
        """
        for index, axis in enumerate(self.axes):
            if name == axis.name:
                return index

        raise ValueError(f"The axis name {name} could not be found")

    def _to_uhi_(self) -> dict[str, Any]:
        from .serialization import to_uhi

        return to_uhi(self)

    @classmethod
    def from_columns(
        cls,
        data: Mapping[str, ArrayLike],
        axes: Sequence[str | AxisProtocol],
        *,
        weight: str | None = None,
        storage: hist.storage.Storage = hist.storage.Double(),  # noqa: B008
    ) -> Self:
        axes_list: list[Any] = []
        for ax in axes:
            if isinstance(ax, str):
                assert ax in data, f"{ax} must be present in data={list(data)}"
                cats = set(data[ax])
                if all(isinstance(a, str) for a in cats):
                    axes_list.append(hist.axis.StrCategory(sorted(cats), name=ax))
                elif all(isinstance(a, int) for a in cats):
                    axes_list.append(hist.axis.IntCategory(sorted(cats), name=ax))
                else:
                    raise TypeError(
                        f"{ax} must be all int or strings if axis not given"
                    )
            elif not ax.name or ax.name not in data:
                raise TypeError("All axes must have names present in the data")
            else:
                axes_list.append(ax)

        weight_arr = data[weight] if weight else None

        self = cls(*axes_list, storage=storage)
        data_list = {x.name: data[x.name] for x in axes_list}
        self.fill(**data_list, weight=weight_arr)
        return self

    def project(self, *args: int | str) -> Self | float | bh.accumulators.Accumulator:
        """
        Projection of axis idx.
        """
        int_args = [self._name_to_index(a) if isinstance(a, str) else a for a in args]
        return super().project(*int_args)

    @property
    def T(self) -> Self:
        return self.project(*reversed(range(self.ndim)))  # type: ignore[return-value]

    def fill(
        self,
        *args: ArrayLike,
        weight: ArrayLike | None = None,
        sample: ArrayLike | None = None,
        threads: int | None = None,
        **kwargs: ArrayLike,
    ) -> Self:
        """
        Insert data into the histogram using names and indices, return
        a Hist object.
        """

        if (
            issubclass(self.storage_type, (bh.storage.Mean, bh.storage.WeightedMean))
            and sample is None
        ):
            msg = "A Mean-based storage requires a sample= argument with the values to average"
            raise TypeError(msg)

        data_dict = {
            self._name_to_index(k) if isinstance(k, str) else k: v  # type: ignore[redundant-expr]
            for k, v in kwargs.items()
        }

        if set(data_dict) != set(range(len(args), self.ndim)):
            expected = set(range(len(args), self.ndim))
            provided = set(data_dict)

            missing = expected - provided
            missing_values = [self.axes[i].name for i in missing]

            raise TypeError(
                f"Missing values for single/multiple axis : {missing_values}."
                f"Expected axes : {[ax.name for ax in self.axes]}"
            )

        data = (data_dict[i] for i in range(len(args), self.ndim))

        # Quick check for a common mistake with a weighted fill
        if (
            weight is not None
            and args
            and hasattr(weight, "__len__")
            and hasattr(args[0], "__len__")
            and len(weight) != len(args[0])
        ):
            raise ValueError(
                "Weight array must match the length of the input data for a given data"
            )

        return super().fill(*args, *data, weight=weight, sample=sample, threads=threads)

    def fill_flattened(
        self: Self,
        *args: Any,
        weight: Any | None = None,
        sample: Any | None = None,
        threads: int | None = None,
        **kwargs: Any,
    ) -> Self:
        axis_names = {ax.name for ax in self.axes}

        non_user_kwargs = {}
        if weight is not None:
            non_user_kwargs["weight"] = weight
        if sample is not None:
            non_user_kwargs["sample"] = sample

        # Single arguments are either arrays for single-dimensional histograms, or
        # structures for multi-dimensional hists that must first be unpacked
        if len(args) == 1 and not kwargs:
            (arg,) = args
            destructured = interop.destructure(arg)
            # Try to unpack the array, if it's valid to do so, i.e. is the Awkward Array a record array?
            if destructured is None:
                # Can't unpack, fall back on broadcasting single array (to flatten and convert)
                broadcast = interop.broadcast_and_flatten(
                    (arg, *non_user_kwargs.values())
                )
                # Partition out non-user args
                user_args_broadcast = broadcast[:1]
                user_kwargs_broadcast = {}
                non_user_kwargs_broadcast = dict(
                    zip(non_user_kwargs.keys(), broadcast[1:])
                )
            else:
                # Result must be broadcast, so unpack and rebuild
                broadcast = interop.broadcast_and_flatten(
                    (*destructured.values(), *non_user_kwargs.values())
                )
                # Partition into user and non-user args
                user_args_broadcast = ()
                user_kwargs_broadcast = {
                    k: v
                    for k, v in zip(destructured, broadcast[: len(destructured)])
                    if k in axis_names
                }
                non_user_kwargs_broadcast = dict(
                    zip(non_user_kwargs, broadcast[len(destructured) :])
                )
        # Multiple args: broadcast and flatten!
        else:
            inputs = (*args, *kwargs.values(), *non_user_kwargs.values())
            broadcast = interop.broadcast_and_flatten(inputs)
            user_args_broadcast = broadcast[: len(args)]
            user_kwargs_broadcast = dict(
                zip(kwargs, broadcast[len(args) : len(args) + len(kwargs)])
            )
            non_user_kwargs_broadcast = dict(
                zip(non_user_kwargs, broadcast[len(args) + len(kwargs) :])
            )
        return self.fill(
            *user_args_broadcast,
            **user_kwargs_broadcast,
            threads=threads,
            **non_user_kwargs_broadcast,
        )

    def sort(
        self,
        axis: int | str,
        key: (
            Callable[[int], SupportsLessThan] | Callable[[str], SupportsLessThan] | None
        ) = None,
        reverse: bool = False,
    ) -> Self:
        """
        Sort a categorical axis.
        """

        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            sorted_cats = sorted(self.axes[axis], key=key, reverse=reverse)
            # This can only return Self, not float, etc., so we ignore the extra types here
            return self[{axis: [bh.loc(x) for x in sorted_cats]}]  # type: ignore[dict-item, return-value]

    def _convert_index_wildcards(self, x: Any, ax_id: str | int | None = None) -> Any:
        """
        Convert wildcards to available indices before passing to bh.loc
        """

        if not any(
            isinstance(x, t) for t in [str, list]
        ):  # Process only lists and strings
            return x
        _x = x if isinstance(x, list) else [x]  # Convert to list if not already
        if not all(isinstance(n, str) for n in _x):
            return x
        if any(any(special in pattern for special in ["*", "?"]) for pattern in _x):
            available = [n for n in self.axes[ax_id] if isinstance(n, str)]
            all_matches = []
            for pattern in _x:
                all_matches.append(
                    [k for k in available if fnmatch.fnmatch(k, pattern)]
                )
            matches = list(
                dict.fromkeys(list(itertools.chain.from_iterable(all_matches)))
            )
            if len(matches) == 0:
                raise ValueError(f"No matches found for {x}")
            return matches
        return x

    def _loc_shortcut(self, x: Any, ax_id: str | int | None = None) -> Any:
        """
        Convert some specific indices to location.
        """
        x = self._convert_index_wildcards(x, ax_id)
        if isinstance(x, list):
            return [self._loc_shortcut(each) for each in x]
        if isinstance(x, slice):
            return slice(
                self._loc_shortcut(x.start),
                self._loc_shortcut(x.stop),
                self._step_shortcut(x.step),
            )
        if isinstance(x, complex):
            if x.real % 1 != 0:
                raise ValueError("The real part should be an integer")
            return bh.loc(x.imag, int(x.real))
        if isinstance(x, str):
            return bh.loc(x)
        return x

    @staticmethod
    def _step_shortcut(x: Any) -> Any:
        """
        Convert some specific indices to step.
        """

        if not isinstance(x, complex):
            return x

        if x.real != 0:
            raise ValueError("The step should not have real part")
        if x.imag % 1 != 0:
            raise ValueError("The imaginary part should be an integer")
        return bh.rebin(int(x.imag))

    def _index_transform(self, index: list[IndexingExpr] | IndexingExpr) -> Any:
        """
        Auxiliary function for __getitem__ and __setitem__.
        """

        if isinstance(index, dict):
            new_indices = {
                (
                    self._name_to_index(k) if isinstance(k, str) else k
                ): self._loc_shortcut(v, k)
                for k, v in index.items()
            }
            if len(new_indices) != len(index):
                raise ValueError(
                    "Duplicate index keys, numbers and names cannot overlap"
                )
            return new_indices

        if not isinstance(index, tuple):
            index = (index,)  # type: ignore[assignment]

        return tuple(self._loc_shortcut(v, i) for i, (v) in enumerate(index))  # type: ignore[arg-type]

    def __getitem__(  # type: ignore[override]
        self, index: IndexingExpr
    ) -> Self | float | bh.accumulators.Accumulator:
        """
        Get histogram item.
        """

        return super().__getitem__(self._index_transform(index))

    def __setitem__(  # type: ignore[override]
        self, index: IndexingExpr, value: ArrayLike | bh.accumulators.Accumulator
    ) -> None:
        """
        Set histogram item.
        """

        return super().__setitem__(self._index_transform(index), value)

    def profile(self, axis: int | str) -> Self:
        """
        Returns a profile (Mean/WeightedMean) histogram from a normal histogram
        with N-1 axes. The axis given is profiled over and removed from the
        final histogram.
        """

        if self.kind != bh.Kind.COUNT:
            raise TypeError("Profile requires a COUNT histogram")

        axes = list(self.axes)
        iaxis = axis if isinstance(axis, int) else self._name_to_index(axis)
        axes.pop(iaxis)

        values = self.values()
        tmp_variances = self.variances()
        variances = tmp_variances if tmp_variances is not None else values
        centers = self.axes[iaxis].centers

        count = np.sum(values, axis=iaxis)

        num = np.tensordot(values, centers, ([iaxis], [0]))
        num_err = np.sqrt(np.tensordot(variances, centers**2, ([iaxis], [0])))

        den = np.sum(values, axis=iaxis)
        den_err = np.sqrt(np.sum(variances, axis=iaxis))

        with np.errstate(invalid="ignore"):
            new_values = num / den
            new_variances = (num_err / den) ** 2 - (den_err * num / den**2) ** 2

        retval = self.__class__(*axes, storage=hist.storage.Mean())
        retval[...] = np.stack([count, new_values, count * new_variances], axis=-1)
        return retval

    def density(self) -> np.typing.NDArray[Any]:
        """
        Density NumPy array.
        """
        total = np.sum(self.values()) * functools.reduce(operator.mul, self.axes.widths)
        dens: np.typing.NDArray[Any] = self.values() / np.where(total > 0, total, 1)
        return dens

    def show(self, **kwargs: Any) -> Any:
        """
        Pretty print histograms to the console.
        """

        return histoprint.print_hist(self, **kwargs)

    def plot(
        self, *args: Any, overlay: str | None = None, **kwargs: Any
    ) -> Hist1DArtists | Hist2DArtists:
        """
        Plot method for BaseHist object.
        """
        ordered_traits = sum(t.ordered for t in self.axes.traits)
        discrete_traits = sum(t.discrete for t in self.axes.traits)
        _has_categorical = ordered_traits == discrete_traits == 1
        _project = _has_categorical or overlay is not None

        if self.ndim == 1 or (self.ndim == 2 and _project):
            return self.plot1d(*args, overlay=overlay, **kwargs)

        if self.ndim == 2:
            return self.plot2d(*args, **kwargs)

        raise NotImplementedError("Please project to 1D or 2D before calling plot")

    def plot1d(
        self,
        *,
        ax: matplotlib.axes.Axes | None = None,
        overlay: str | int | None = None,
        legend: bool = True,
        **kwargs: Any,
    ) -> Hist1DArtists:
        """
        Plot1d method for BaseHist object.

        Parameters
        ----------
        ax : matplotlib.axes.Axes, optional
            Axes to plot on. If None, uses current axes or creates new ones.
        overlay : str or int, optional
            Name or index of the axis to overlay. If None, automatically selects
            the first discrete axis for multi-dimensional histograms.
        legend : bool, default True
            Whether to automatically add a legend when plotting stacked categories.
            The legend title is set from the axis label if available.
        **kwargs : Any
            Additional keyword arguments passed to the underlying plot functions.

        Returns
        -------
        Hist1DArtists
            The matplotlib artists created by the plot.
        """

        from hist import plot

        if self.ndim == 1:
            return plot.histplot(self, ax=ax, **_proc_kw_for_lw(kwargs))
        if overlay is None:
            (overlay,) = (i for i, ax in enumerate(self.axes) if ax.traits.discrete)
        assert overlay is not None
        cat_ax = self.axes[overlay]
        icats = np.arange(len(cat_ax))
        cats = cat_ax if cat_ax.traits.discrete else icats
        d1hists = [self[{overlay: cat}] for cat in icats]
        if "label" in kwargs:
            if not isinstance(kwargs["label"], str) and len(kwargs["label"]) == len(
                cats
            ):
                cats = kwargs["label"]
                kwargs.pop("label")
            elif isinstance(kwargs["label"], str):
                cats = [kwargs["label"]] * len(cats)
                kwargs.pop("label")
            else:
                raise ValueError(
                    f"label ``{kwargs['label']}`` not understood for {len(cats)} categories"
                )
        artists = plot.histplot(d1hists, ax=ax, label=cats, **_proc_kw_for_lw(kwargs))
        if legend:
            # Try to set legend title from axis label if available
            if ax is None:
                # Get axis from the first artist (mplhep returns Hist1DArtists tuple)
                # This will raise an error if artists is empty or doesn't have the expected structure,
                # which is intended behavior as specified in the requirements
                ax = artists[0].stairs.axes
            handles, _ = ax.get_legend_handles_labels()
            if handles:
                title = getattr(cat_ax, "label", None)
                ax.legend(title=title if title else None)
        return artists

    def plot2d(
        self,
        *,
        ax: matplotlib.axes.Axes | None = None,
        **kwargs: Any,
    ) -> Hist2DArtists:
        """
        Plot2d method for BaseHist object.
        """

        from hist import plot

        return plot.hist2dplot(self, ax=ax, **_proc_kw_for_lw(kwargs))

    def plot2d_full(
        self,
        *,
        ax_dict: dict[str, matplotlib.axes.Axes] | None = None,
        **kwargs: Any,
    ) -> tuple[Hist2DArtists, Hist1DArtists, Hist1DArtists]:
        """
        Plot2d_full method for BaseHist object.

        Pass a dict of axes to ``ax_dict``, otherwise, the current figure will be used.
        """
        # Type judgement

        from hist import plot

        return plot.plot2d_full(self, ax_dict=ax_dict, **kwargs)

    def plot_ratio(
        self,
        other: hist.BaseHist
        | Callable[[np.typing.NDArray[Any]], np.typing.NDArray[Any]]
        | str,
        *,
        ax_dict: dict[str, matplotlib.axes.Axes] | None = None,
        **kwargs: Any,
    ) -> tuple[MainAxisArtists, RatiolikeArtists]:
        """
        ``plot_ratio`` method for ``BaseHist`` object.

        Return a tuple of artists following a structure of
        ``(main_ax_artists, subplot_ax_artists)``
        """

        from hist import plot

        return plot._plot_ratiolike(
            self, other, ax_dict=ax_dict, view="ratio", **kwargs
        )

    def plot_pull(
        self,
        func: Callable[[np.typing.NDArray[Any]], np.typing.NDArray[Any]] | str,
        *,
        ax_dict: dict[str, matplotlib.axes.Axes] | None = None,
        **kwargs: Any,
    ) -> tuple[FitResultArtists, RatiolikeArtists]:
        """
        ``plot_pull`` method for ``BaseHist`` object.

        Return a tuple of artists following a structure of
        ``(main_ax_artists, subplot_ax_artists)``
        """

        from hist import plot

        return plot._plot_ratiolike(self, func, ax_dict=ax_dict, view="pull", **kwargs)

    def plot_pie(
        self,
        *,
        ax: matplotlib.axes.Axes | None = None,
        **kwargs: Any,
    ) -> Any:
        from hist import plot

        return plot.plot_pie(self, ax=ax, **kwargs)

    def stack(self, axis: int | str) -> hist.stack.Stack:
        """
        Returns a stack from a normal histogram axes.
        """
        if self.ndim < 2:
            raise RuntimeError("Cannot stack with less than two axis")
        stack_histograms: Iterator[BaseHist] = [  # type: ignore[assignment]
            self[{axis: i}] for i in range(len(self.axes[axis]))
        ]
        for name, h in zip(self.axes[axis], stack_histograms):
            h.name = name

        return hist.stack.Stack(*stack_histograms)

    def integrate(
        self,
        name: int | str,
        i_or_list: InnerIndexing | list[str | int] | None = None,
        j: InnerIndexing | None = None,
    ) -> Self | float | bh.accumulators.Accumulator:
        if isinstance(i_or_list, list):
            with warnings.catch_warnings():
                warnings.simplefilter("ignore")

                # TODO: We could teach the typing system that list always returns Self type
                selection: Self = self[{name: i_or_list}]  # type: ignore[assignment, dict-item]
                return selection[{name: slice(0, len(i_or_list), sum)}]

        return self[{name: slice(i_or_list, j, sum)}]

    def sum(self, flow: bool = False) -> float | bh.accumulators.Accumulator:
        """
        Compute the sum over the histogram bins (optionally including the flow bins).
        """
        # TODO: This method will go away once we can guarantee a modern boost-histogram (1.3.2 or better)
        if any(x == 0 for x in (self.axes.extent if flow else self.axes.size)):
            storage = self.storage_type
            if issubclass(storage, (bh.storage.AtomicInt64, bh.storage.Int64)):
                return 0
            if issubclass(storage, (bh.storage.Double, bh.storage.Unlimited)):
                return 0.0
            if issubclass(storage, bh.storage.Weight):
                return bh.accumulators.WeightedSum(0, 0)
            if issubclass(storage, bh.storage.Mean):
                return bh.accumulators.Mean(0, 0, 0)
            if issubclass(storage, bh.storage.WeightedMean):
                return bh.accumulators.WeightedMean(0, 0, 0, 0)
            raise AssertionError(f"Unsupported storage type {storage}")

        return super().sum(flow=flow)



from __future__ import annotations

from typing import Any, Literal

import numpy as np

try:
    from scipy import stats
except ModuleNotFoundError:
    from sys import stderr

    print(  # noqa: T201
        "hist.intervals requires scipy. Please install hist[plot] or manually install scipy.",
        file=stderr,
    )
    raise

__all__ = ("clopper_pearson_interval", "poisson_interval", "ratio_uncertainty")


def __dir__() -> tuple[str, ...]:
    return __all__


def poisson_interval(
    values: np.typing.NDArray[Any],
    variances: np.typing.NDArray[Any] | None = None,
    coverage: float | None = None,
) -> np.typing.NDArray[Any]:
    r"""
    The Frequentist coverage interval for Poisson-distributed observations.

    What is calculated is the "Garwood" interval, c.f.
    `V. Patil, H. Kulkarni (Revstat, 2012) <https://www.ine.pt/revstat/pdf/rs120203.pdf>`_
    or http://ms.mcmaster.ca/peter/s743/poissonalpha.html.
    If ``variances`` is supplied, the data is assumed to be weighted, and the
    unweighted count is approximated by ``values**2/variances``, which effectively
    scales the unweighted Poisson interval by the average weight.
    This may not be the optimal solution: see
    `10.1016/j.nima.2014.02.021 <https://doi.org/10.1016/j.nima.2014.02.021>`_
    (`arXiv:1309.1287 <https://arxiv.org/abs/1309.1287>`_) for a proper treatment.

    In cases where the value is zero, an upper limit is well-defined only in the case of
    unweighted data, so if ``variances`` is supplied, the upper limit for a zero value
    will be set to ``NaN``.

    Args:
        values: Sum of weights.
        variances: Sum of weights squared.
        coverage: Central coverage interval.
          Default is one standard deviation, which is roughly ``0.68``.

    Returns:
        The Poisson central coverage interval.
    """
    # Parts originally contributed to coffea
    # https://github.com/CoffeaTeam/coffea/blob/8c58807e199a7694bf15e3803dbaf706d34bbfa0/LICENSE
    if coverage is None:
        coverage = stats.norm.cdf(1) - stats.norm.cdf(-1)
    if variances is None:
        interval_min = stats.chi2.ppf((1 - coverage) / 2, 2 * values) / 2.0
        interval_min[values == 0.0] = 0.0  # chi2.ppf produces NaN for values=0
        interval_max = stats.chi2.ppf((1 + coverage) / 2, 2 * (values + 1)) / 2.0
    else:
        scale = np.ones_like(values)
        mask = np.isfinite(values) & (values != 0)
        np.divide(variances, values, out=scale, where=mask)
        counts: np.typing.NDArray[Any] = values / scale
        interval_min = scale * stats.chi2.ppf((1 - coverage) / 2, 2 * counts) / 2.0
        interval_min[values == 0.0] = 0.0  # chi2.ppf produces NaN for values=0
        interval_max = (
            scale * stats.chi2.ppf((1 + coverage) / 2, 2 * (counts + 1)) / 2.0
        )
        interval_max[values == 0.0] = np.nan
    return np.stack((interval_min, interval_max))


def clopper_pearson_interval(
    num: np.typing.NDArray[Any],
    denom: np.typing.NDArray[Any],
    coverage: float | None = None,
) -> np.typing.NDArray[Any]:
    r"""
    Compute the Clopper-Pearson coverage interval for a binomial distribution.
    c.f. http://en.wikipedia.org/wiki/Binomial_proportion_confidence_interval

    Args:
        num: Numerator or number of successes.
        denom: Denominator or number of trials.
        coverage: Central coverage interval.
          Default is one standard deviation, which is roughly ``0.68``.

    Returns:
        The Clopper-Pearson central coverage interval.
    """
    # Parts originally contributed to coffea
    # https://github.com/CoffeaTeam/coffea/blob/8c58807e199a7694bf15e3803dbaf706d34bbfa0/LICENSE
    if coverage is None:
        coverage = stats.norm.cdf(1) - stats.norm.cdf(-1)
    # Numerator is subset of denominator
    if np.any(num > denom):
        raise ValueError(
            "Found numerator larger than denominator while calculating binomial uncertainty"
        )
    interval_min = stats.beta.ppf((1 - coverage) / 2, num, denom - num + 1)
    interval_max = stats.beta.ppf((1 + coverage) / 2, num + 1, denom - num)
    interval = np.stack((interval_min, interval_max))
    interval[0, num == 0.0] = 0.0
    interval[1, num == denom] = 1.0
    return interval


def ratio_uncertainty(
    num: np.typing.NDArray[Any],
    denom: np.typing.NDArray[Any],
    uncertainty_type: Literal["poisson", "poisson-ratio", "efficiency"] = "poisson",
) -> Any:
    r"""
    Calculate the uncertainties for the values of the ratio ``num/denom`` using
    the specified coverage interval approach.

    Args:
        num: Numerator or number of successes.
        denom: Denominator or number of trials.
        uncertainty_type: Coverage interval type to use in the calculation of
         the uncertainties.

         * ``"poisson"`` (default) implements the Garwood confidence interval for
           a Poisson-distributed numerator scaled by the denominator.
           See :func:`hist.intervals.poisson_interval` for further details.
         * ``"poisson-ratio"`` implements a confidence interval for the ratio ``num / denom``
           assuming it is an estimator of the ratio of the expected rates from
           two independent Poisson distributions.
           It over-covers to a similar degree as the Clopper-Pearson interval
           does for the Binomial efficiency parameter estimate.
         * ``"efficiency"`` implements the Clopper-Pearson confidence interval
           for the ratio ``num / denom`` assuming it is an estimator of a Binomial
           efficiency parameter.
           This is only valid if the entries contributing to ``num`` are a strict
           subset of those contributing to ``denom``.

    Returns:
        The uncertainties for the ratio.
    """
    # Note: As return is a numpy ufuncs the type is "Any"
    with np.errstate(divide="ignore", invalid="ignore"):
        # Nota bene: x/0 = inf, 0/0 = nan
        ratio = num / denom
    if uncertainty_type == "poisson":
        with np.errstate(divide="ignore", invalid="ignore"):
            ratio_variance = num * np.power(denom, -2.0)
        ratio_uncert = np.abs(poisson_interval(ratio, ratio_variance) - ratio)
    elif uncertainty_type == "poisson-ratio":
        # Details: see https://github.com/scikit-hep/hist/issues/279
        p_lim = clopper_pearson_interval(num, num + denom)
        with np.errstate(divide="ignore", invalid="ignore"):
            r_lim: np.typing.NDArray[Any] = p_lim / (1 - p_lim)
            ratio_uncert = np.abs(r_lim - ratio)
    elif uncertainty_type == "efficiency":
        ratio_uncert = np.abs(clopper_pearson_interval(num, denom) - ratio)
    else:
        raise TypeError(
            f"'{uncertainty_type}' is an invalid option for uncertainty_type."
        )
    return ratio_uncert



from __future__ import annotations

from typing import Protocol

from ._compat.typing import Self


class SupportsStr(Protocol):
    def __str__(self) -> str: ...


class XML:
    def __init__(self, *contents: XML | SupportsStr, **kargs: SupportsStr) -> None:
        self.properties = kargs
        self.contents = contents

    @property
    def name(self) -> str:
        return self.__class__.__name__

    @property
    def start(self) -> str:
        if self.properties:
            ending = " " + " ".join(
                f'{a.replace("_", "-")}="{b}"' for a, b in self.properties.items()
            )
        else:
            ending = ""
        return self.name + ending

    def __str__(self) -> str:
        if not self.contents:
            return f"<{self.start}/>"

        contents = "\n".join(str(s) for s in self.contents)
        return f"<{self.start}>\n{contents}\n</{self.name}>"


class html(XML):
    def _repr_xml_(self) -> str:
        return str(self)


class svg(XML):
    def __init__(self, *args: XML | str, **kwargs: str) -> None:
        super().__init__(*args, xmlns="http://www.w3.org/2000/svg", **kwargs)

    def _repr_svg_(self) -> str:
        return str(self)


class circle(XML):
    pass


class line(XML):
    pass


class p(XML):
    pass


class div(XML):
    pass


class rect(XML):
    @classmethod
    def pad(
        cls,
        x: float,
        y: float,
        scale_x: float,
        scale_y: float,
        height: float,
        left_edge: float,
        right_edge: float,
        pad_x: float = 0,
        pad_y: float = 0,
        opacity: float = 1,
        stroke_width: float = 2,
    ) -> Self:
        width = right_edge - left_edge
        top_y = y

        height *= 1 - 2 * pad_y
        top_y = top_y * (1 - 2 * pad_y) + pad_y
        width = width * (1 - 2 * pad_x)
        color = int(opacity * 255)

        return cls(
            x=x * scale_x,
            y=top_y * scale_y,
            width=width * scale_x,
            height=height * scale_y,
            style=f"fill:rgb({color}, {color}, {color});stroke:rgb(0,0,0);stroke-width:{stroke_width}",
        )


class ellipse(XML):
    pass


class polygon(XML):
    pass


class polyline(XML):
    pass


# path
class text(XML):
    pass


class g(XML):
    pass



from __future__ import annotations

import typing
from collections.abc import Iterable
from typing import Any, Protocol

import boost_histogram.axis as bha

import hist

from ..axestuple import ArrayTuple, NamedAxesTuple
from . import transform

__all__ = (
    "ArrayTuple",
    "AxesMixin",
    "AxisProtocol",
    "Boolean",
    "IntCategory",
    "Integer",
    "NamedAxesTuple",
    "Regular",
    "StrCategory",
    "Variable",
    "transform",
)


def __dir__() -> tuple[str, ...]:
    return __all__


class CoreAxisProtocol(Protocol):
    metadata: dict[str, Any]  # boost-histogram < 1.6
    raw_metadata: dict[str, Any]


class AxisProtocol(Protocol):
    @property
    def name(self) -> str: ...

    label: str
    _ax: CoreAxisProtocol

    @property
    def _raw_metadata(self) -> dict[str, Any]: ...


class AxesMixin:
    __slots__ = ()

    # Support mixing before or after a bh class
    def __init_subclass__(cls, **kwargs: Any) -> None:
        super().__init_subclass__(**kwargs)

    @property
    def _raw_metadata(self: AxisProtocol) -> dict[str, Any]:
        # boost-histogram < 1.6
        if hasattr(self._ax, "metadata"):
            return self._ax.metadata
        return self._ax.raw_metadata

    @property
    def name(self: AxisProtocol) -> str:
        """
        Get the name for the Regular axis
        """
        return typing.cast(str, self._raw_metadata.get("name", ""))

    @property
    def label(self: AxisProtocol) -> str:
        """
        Get or set the label for the Regular axis
        """
        return self._raw_metadata.get("label", "") or self.name

    @label.setter
    def label(self: AxisProtocol, value: str) -> None:
        self._raw_metadata["label"] = value

    def _repr_args_(self: AxisProtocol) -> list[str]:
        """
        Return options for use in repr.
        """
        ret: list[str] = super()._repr_args_()  # type: ignore[misc]

        if self.name:
            ret.append(f"name={self.name!r}")
        if self.label and self.label != self.name:
            ret.append(f"label={self.label!r}")

        return ret


class Regular(AxesMixin, bha.Regular, family=hist):
    __slots__ = ()

    def __init__(
        self,
        bins: int,
        start: float,
        stop: float,
        *,
        name: str = "",
        label: str = "",
        metadata: Any = None,
        flow: bool = True,
        underflow: bool | None = None,
        overflow: bool | None = None,
        growth: bool = False,
        circular: bool = False,
        # pylint: disable-next=redefined-outer-name
        transform: bha.transform.AxisTransform | None = None,
        __dict__: dict[str, Any] | None = None,
    ) -> None:
        super().__init__(
            bins,
            start,
            stop,
            metadata=metadata,
            underflow=flow if underflow is None else underflow,
            overflow=flow if overflow is None else overflow,
            growth=growth,
            circular=circular,
            transform=transform,
            __dict__=__dict__,
        )
        self._raw_metadata["name"] = name
        self.label: str = label


class Boolean(AxesMixin, bha.Boolean, family=hist):
    __slots__ = ()

    def __init__(
        self,
        *,
        name: str = "",
        label: str = "",
        metadata: Any = None,
        __dict__: dict[str, Any] | None = None,
    ) -> None:
        super().__init__(
            metadata=metadata,
            __dict__=__dict__,
        )
        self._raw_metadata["name"] = name
        self.label: str = label


class Variable(AxesMixin, bha.Variable, family=hist):
    __slots__ = ()

    def __init__(
        self,
        edges: Iterable[float],
        *,
        name: str = "",
        label: str = "",
        metadata: Any = None,
        flow: bool = True,
        underflow: bool | None = None,
        overflow: bool | None = None,
        growth: bool = False,
        circular: bool = False,
        __dict__: dict[str, Any] | None = None,
    ) -> None:
        super().__init__(
            edges,
            metadata=metadata,
            underflow=flow if underflow is None else underflow,
            overflow=flow if overflow is None else overflow,
            growth=growth,
            circular=circular,
            __dict__=__dict__,
        )
        self._raw_metadata["name"] = name
        self.label: str = label


class Integer(AxesMixin, bha.Integer, family=hist):
    __slots__ = ()

    def __init__(
        self,
        start: int,
        stop: int,
        *,
        name: str = "",
        label: str = "",
        metadata: Any = None,
        flow: bool = True,
        underflow: bool | None = None,
        overflow: bool | None = None,
        growth: bool = False,
        circular: bool = False,
        __dict__: dict[str, Any] | None = None,
    ) -> None:
        super().__init__(
            start,
            stop,
            metadata=metadata,
            underflow=flow if underflow is None else underflow,
            overflow=flow if overflow is None else overflow,
            growth=growth,
            circular=circular,
            __dict__=__dict__,
        )
        self._raw_metadata["name"] = name
        self.label: str = label


class IntCategory(AxesMixin, bha.IntCategory, family=hist):
    __slots__ = ()

    def __init__(
        self,
        categories: Iterable[int],
        *,
        name: str = "",
        label: str = "",
        metadata: Any = None,
        growth: bool = False,
        flow: bool = True,
        overflow: bool | None = None,
        __dict__: dict[str, Any] | None = None,
    ) -> None:
        has_flow = flow if overflow is None else overflow
        super().__init__(
            categories,
            metadata=metadata,
            growth=growth,
            overflow=has_flow,
            __dict__=__dict__,
        )
        self._raw_metadata["name"] = name
        self.label: str = label


class StrCategory(AxesMixin, bha.StrCategory, family=hist):
    __slots__ = ()

    def __init__(
        self,
        categories: Iterable[str],
        *,
        name: str = "",
        label: str = "",
        metadata: Any = None,
        growth: bool = False,
        flow: bool = True,
        overflow: bool | None = None,
        __dict__: dict[str, Any] | None = None,
    ) -> None:
        has_flow = flow if overflow is None else overflow
        super().__init__(
            categories,
            metadata=metadata,
            growth=growth,
            overflow=has_flow,
            __dict__=__dict__,
        )
        self._raw_metadata["name"] = name
        self.label: str = label



from __future__ import annotations

import copy
from typing import Any, TypeVar

from boost_histogram import Histogram

from .. import Hist, __version__

__all__ = ["from_uhi", "remove_writer_info", "to_uhi"]


def from_uhi(data: dict[str, Any], /) -> Hist:
    import boost_histogram.serialization as bhs

    return Hist(bhs.from_uhi(data))


def to_uhi(h: Histogram, /) -> dict[str, Any]:
    import boost_histogram.serialization as bhs

    d = bhs.to_uhi(h)
    d["writer_info"]["hist"] = {"version": __version__}
    return d


T = TypeVar("T", bound="dict[str, Any]")


def remove_writer_info(obj: T) -> T:
    """Removes all hist writer_info from a histogram dict, axes dict, or storage dict. Makes copies where required, and the outer dictionary is always copied."""

    obj = copy.copy(obj)
    if "hist" in obj.get("writer_info", {}):
        obj["writer_info"] = copy.copy(obj["writer_info"])
        del obj["writer_info"]["hist"]

    if "axes" in obj:
        obj["axes"] = [remove_writer_info(ax) for ax in obj["axes"]]
    if "storage" in obj:
        obj["storage"] = remove_writer_info(obj["storage"])

    return obj
